%INTEGRATE_RANGE_NEXUS  Azimuthal integration of SAXS patterns from NeXus
%   or H5 files
%
% beamline.integrate_range_nexus(startScanNr, varargin)
%
% Inputs
% ** startScanNr    Scan number from which to start
%
% *optional* (parameter, value) pairs
% ** ('stopScanNr', scanNum)    Final scan number, default = inf
% ** ('detector','pilatus_1')   Detector group inside the H5 file, default = 'pilatus_1'
% ** ('outdirData','path/')     Output path default = '~/Data10/analysis/radial_integration/'
% ** ('integMask', 'path/filename') Azimuthal integration mask, default = '~/Data10/analysis/data/integration_masks.mat'
% ** ('basePath', '~/Data10')   Base path to provide to utils.find_nexus_file, default = '~/Data10'     
%
% see also: utils.find_nexus_file, beamline.radial_integ
%
% EXAMPLES:
%   beamline.integrate_range_nexus(5);
%   beamline.integrate_range_nexus(5,'detector','pilatus_2');


%*-----------------------------------------------------------------------*
%|                                                                       |
%|  Except where otherwise noted, this work is licensed under a          |
%|  Creative Commons Attribution-NonCommercial-ShareAlike 4.0            |
%|  International (CC BY-NC-SA 4.0) license.                             |
%|                                                                       |
%|  Copyright (c) 2019 by Paul Scherrer Institute (http://www.psi.ch)    |
%|                                                                       |
%|       Author: CXS group, PSI                                          |
%*-----------------------------------------------------------------------*
% You may use this code with the following provisions:
%
% If the code is fully or partially redistributed, or rewritten in another
%   computing language this notice should be included in the redistribution.
%
% If this code, or subfunctions or parts of it, is used for research in a 
%   publication or if it is fully or partially rewritten for another 
%   computing language the authors and institution should be acknowledged 
%   in written form in the publication: “Data processing was carried out 
%   using the “cSAXS software package” developed by the CXS group,
%   Paul Scherrer Institut, Switzerland.” 
%   Variations on the latter text can be incorporated upon discussion with 
%   the CXS group if needed to more specifically reflect the use of the package 
%   for the published work.
%
% A publication that focuses on describing features, or parameters, that
%    are already existing in the code should be first discussed with the
%    authors.
%   
% This code and subroutines are part of a continuous development, they 
%    are provided “as they are” without guarantees or liability on part
%    of PSI or the authors. It is the user responsibility to ensure its 
%    proper use and the correctness of the results.


function integrate_range_nexus(startScanNr, varargin)

par = inputParser;
par.addParameter('stopScanNr', inf, @isnumeric)
par.addParameter('detector', 'pilatus_1', @ischar)
par.addParameter('outdirData', '~/Data10/analysis/radial_integration', @ischar)
par.addParameter('integMask', '~/Data10/analysis/data/integration_masks.mat', @ischar)
par.addParameter('basePath', '~/Data10', @ischar)


par.parse(varargin{:})
vars = par.Results;
vars.startScanNr = startScanNr;

scan = startScanNr;
while scan <= vars.stopScanNr
    fname = utils.find_nexus_file(vars.basePath, scan);
    while ~utils.nexus_file_is_ready(fname)
        fprintf('Waiting for scan %05d.\n', scan);
        pause(5);
        fname = utils.find_nexus_file(vars.basePath, scan);
    end
    
    % read nexus file
    beamline.radial_integ(fname, 'FilenameIntegMasks', vars.integMask, 'OutdirData', vars.outdirData, 'H5Location', fullfile('/entry/instrument/', vars.detector), 'Detector', vars.detector, 'UseNexus', true);
    scan = scan +1;
end





end