#!/usr/bin/env python

# *-----------------------------------------------------------------------*
# |                                                                       |
# |  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
# |                                                                       |
# |              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
# *-----------------------------------------------------------------------*
'''
Parse Header of .cbf-Files and decompress the binary part.
'''

import sys,os,time,re,mmap
import numpy as np
import ctypes as ct
import traceback
import logging
import md5,base64
_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

try:
  if sys.platform == "linux2":
    #sys.path.insert(0,r'libDetXR/python')
    sys.path.insert(0,r'./')
except BaseException  as err:
  _log.critical('Path Error'+str(err))

from libDetXR import LibDetXR

class CbfWriter:
  def __init__(self):
    self.header=header={}
    header['version']='''###CBF: VERSION 1.5, CBFlib v0.7.8 - PILATUS detectors\r
\r
data_e13900_1_00150_00000_00000\r\n\r\n'''

    header['convention']='_array_data.header_convention "SLS_1.0"\r\n'
    header['contents']='''_array_data.header_contents\r
;\r
# Detector: CbfFile.py generated\r
# 2012-03-16T16:37:55.679\r
# Pixel_size 172e-6 m x 172e-6 m\r
# Silicon sensor, thickness 0.000320 m\r
# Exposure_time 0.5000000 s\r
# Exposure_period 0.5000000 s\r
# Tau = 338.8e-09 s\r
# Count_cutoff 737891 counts\r
# Threshold_setting: not set\r
# Gain_setting: not set (vrf = 0.000)\r
# N_excluded_pixels = 0\r
# Excluded_pixels: (nil)\r
# Flat_field: (nil)\r
# Trim_file: (nil)\r
# Image_path: /sls/X12SA/Data10/e13900/pilatus_1/S00000-00999/S00150/\r
;\r\n'''

    header['data']='''_array_data.data\r
;\r
--CIF-BINARY-FORMAT-SECTION--\r
Content-Type: application/octet-stream;\r
     conversions="x-CBF_BYTE_OFFSET"\r
Content-Transfer-Encoding: BINARY\r
X-Binary-Size: {binSz:d}\r
X-Binary-ID: 1\r
X-Binary-Element-Type: "{elType}"\r
X-Binary-Element-Byte-Order: LITTLE_ENDIAN\r
Content-MD5: {md5Str}\r
X-Binary-Number-of-Elements: {arr.size:d}\r
X-Binary-Size-Fastest-Dimension: {arr.shape[1]:d}\r
X-Binary-Size-Second-Dimension: {arr.shape[0]:d}\r
X-Binary-Size-Padding: 4095\r
\r\n'''

  def SetData(self,data):
    self.data=data

  def Write(self,fn):
    LibDetXR()
    header=self.header
    f=open(fn,'wb')
    data=self.data
    outBuf=np.ndarray(data.nbytes,dtype=np.uint8,order='C')
    cmpSz=LibDetXR.CompressCBF(data.ctypes.data, data.nbytes, outBuf.ctypes.data, outBuf.nbytes,data.itemsize)
    md5str=base64.b64encode(md5.new(outBuf[:cmpSz]).digest())

    #if data.dtype==np.uint32
    if data.itemsize==4:
      elType='signed 32-bit integer'
    elif data.itemsize==2:
      elType='signed 16-bit integer'
    else:
      raise TypeError(str(data.dtype))

    f.write(header['version'])
    f.write(header['convention'])
    f.write(header['contents'])
    f.write(header['data'].format(binSz=cmpSz,arr=data,md5Str=md5str,elType=elType))
    f.write(CbfParser.cbfHeaderMarkEnd)
    f.write(outBuf[:cmpSz])
    f.close()
    return cmpSz

class CbfParser:
  cbfHeaderMarkEnd = "\x0C\x1A\x04\xD5"
  @staticmethod
  def SetDisplayMode(color=True):
    CbfParser.separator = 20*'-'
    if color:
      #http://en.wikipedia.org/wiki/ANSI_escape_code
      CbfParser.RESET     = '\033[0m'
      CbfParser.BOLD      = '\033[1m'
      CbfParser.BOLD_BLUE = '\033[1;34m'
      CbfParser.BOLD_RED  = '\033[1;31m'
    else:
      CbfParser.RESET     = ''
      CbfParser.BOLD      = ''
      CbfParser.BOLD_BLUE = ''
      CbfParser.BOLD_RED  = ''

  @staticmethod
  def parsePixel_size(txt):
  #http://docs.python.org/2/library/re.html#simulating-scanf
    #f='[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?'
    f='[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
    res=re.search('('+f+')\s*m\s*x\s*('+f+')\s*(.*)',txt).groups()
    val=map(float,res[:-1])
    return (tuple(val),res[-1])
  @staticmethod
  def parseFloatArray(txt):
  #http://docs.python.org/2/library/re.html#simulating-scanf
    #f='[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?'
    f='[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
    res=re.search('('+f+')[\s,]*('+f+')[\s\)]*(.*)',txt).groups()
    val=map(float,res[:-1])
    return (tuple(val),res[-1])
  @staticmethod
  def parseDetector(txt):
    res=txt.split(',')
    res=map(str.strip,res)
    return res
  @staticmethod
  def parseFloat(txt):
    f='[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
    res=re.search('('+f+')\s*(\S*)',txt)
    if res: res=res.groups()
    else:   return txt
    try:
      if res[1]:
        res=(float(res[0]),res[1])
      else:
        res=float(res[0])
    except BaseException as e:
      print sys._getframe().f_code.co_name+' Error: "'+txt+'"',e
    return res
  @staticmethod
  def parseInt(txt):
    i='[-+]?\d+'
    res=re.search('('+i+')\s*(\S*)',txt).groups()
    try:
      if res[1]:
        res=(int(res[0]),res[1])
      else:
        res=int(res[0])
    except BaseException as e:
      print sys._getframe().f_code.co_name+' Error: "'+txt+'"',e
    return res

  _keywords = {
    #key: (pattern, [value_indeces], type)
    #NON_OPTIONAL_KEYWORDS
    'Detector': parseDetector,
    'Pixel': parsePixel_size,
    'Pixel_size': parsePixel_size,
    'Silicon': parseFloat,
    'Exposure_time': parseFloat,
    'Exposure_period': parseFloat,
    'Tau': parseFloat,
    'Count_cutoff': parseInt,
    'Threshold_setting': parseFloat,
    'Gain_setting': None,
    'N_excluded_pixels': parseInt,
    'Excluded_pixels': None,
    'Flat_field': None,
    'Trim_file': None,
    'Image_path': None,
    #OPTIONAL_KEYWORDS
    'Trim_directory': None,
    'Wavelength': parseFloat,
    'Energy_range': parseFloatArray,
    'Detector_distance': parseFloat,
    'Detector_Voffset': parseFloat,
    'Beam_xy': parseFloatArray,
    #'Beam_x': ('Beam_xy', [1], float),
    #'Beam_y': ('Beam_xy', [2], float),
    'Flux': parseFloat,
    'Filter_transmission': parseFloat,
    'Start_angle': parseFloat,
    'Angle_increment': parseFloat,
    'Detector_2theta': parseFloat,
    'Polarization': parseFloat,
    'Alpha': parseFloat,
    'Kappa': parseFloat,
    'Phi': parseFloat,
    'Phi_increment': parseFloat,
    'Chi': parseFloat,
    'Chi_increment': parseFloat,
    'Oscillation_axis': None,
    'N_oscillations': parseInt,
    'Start_position': parseFloat,
    'Position_increment': parseFloat,
    'Shutter_time': parseFloat,
    }

#cannot parse Trim_directory : # Wavelength 1.0000 A
#cannot parse Energy_range : (0, 0) eV
#cannot parse Oscillation_axis : X, CW
#cannot parse Comment : Pixel_size 172e-6 m x 172e-6 m


  def Open(self,fn):
    if not os.path.exists(fn):
      fn=os.path.expanduser(fn)
    self.fn=fn
    self.fh=fh = os.open(fn, os.O_RDONLY)
    self.data=mmap.mmap(fh, 0,  access=mmap.ACCESS_READ)
    self.header=dict()
    self.FindHeaderEnd()

  def SetData(self,data):
    self.data=data
    self.header=dict()
    self.FindHeaderEnd()

  def Close(self):
    os.close(self.fh)

  def FindHeaderEnd(self):
    data=self.data
    self.cbfHeaderEnd = data.find(CbfParser.cbfHeaderMarkEnd)

  def ParseHeader(self,mode=7):
    """
      Parsing the header
      mode is an ored value:
      mode 0: only raw
      mode 1: only section
      mode 2: parse _array_data.header_contents
      mode 4: parse _array_data.data
    """
    data=self.data
    header=self.header
    txt=data[0:self.cbfHeaderEnd]
    if mode==0:
      header['raw']=txt
      return

    pos=0

    m=re.search(r'###CBF:\s*([^\r]*)',txt[pos:])
    if m:
      header['cbfVersion']= m.group(1)
      pos+=m.end()
    m=re.search(r'data_(.*?)[\r\n]',txt[pos:])
    if m:
      header['dataName']= m.group(1)
      pos+=m.end()

    reExpr=re.compile(r'_array_data\.(\S*)(.*?)[\r\n]')

    while(True):
      m=reExpr.search(txt[pos:])
      if m:
        key=m.group(1)
        val=m.group(2).strip()
        if val:
          header[key]=val
          pos+=m.end()
        else:
          pos+=m.end()
          n=re.search(r'(\n;)(.*?)(\1|$)',txt[pos:],re.DOTALL)
          header[key]=n.group(2)
      else:
        break
    if mode|2:
      self.ParseHeaderContents()
    if mode|4:
      self.ParseHeaderData()
    pass

  def ParseHeaderContents(self):
    try:
      txt=self.header['header_contents']
    except KeyError as e:
      #first call parseHeader()
      raise e
    self.header['header_contents']=parsed=dict()

    pos=0
    reExpr=re.compile(r'\s*#+\s*(\w*)([\s:=,]*)(.*?)[\r\n]+')
    while(True):
      if len(parsed)==1:
        #assume date string
        m=re.match(r'\s*#+\s*(.*?)[\r\n]',txt[pos:])
        if m:
          date=m.group(1)
          #try to parse the date:
          p=date.rfind('.')
          for fmt in('%Y-%m-%dT%H:%M:%S','%Y-%b-%dT%H:%M:%S','%Y/%b/%d %H:%M:%S'):
            try:
              e=None
              t=time.strptime(date[:p],fmt)
              ms=int(date[p+1:])
              date=(t,ms)
              break
            except BaseException as e:
              continue
          if e:
            traceback.print_exc(1)
            sys.stderr.write('Date conversion error')
          parsed['date']=date
          pos+=m.end()
          continue
      m=reExpr.match(txt[pos:])
      if m:
        key=m.group(1)
        val=m.group(3).strip()
        #try:
        #  ptrn=CbfParser._translate[key]
        #except KeyError as e:
        #  pass
        #else:
        #  do translation of key
        try:
          kw=CbfParser._keywords[key]
          if type(kw)==staticmethod:
            val=kw.__func__(val)
        except KeyError as e:
          R=CbfParser.RESET; BR=CbfParser.BOLD_RED
          etype, value, tb = sys.exc_info(); filename, lineno, name, line =traceback.extract_tb(tb,1)[0]
          sys.stderr.write('File "%s", line %d, in %s\n' % (filename,lineno,name))
          sys.stderr.write(BR+'cannot parse "%s : %s"\n'%(key,val)+R)
        if val and val!='(nil)':
          parsed[key]=val
        pos+=m.end()
      else:
        break
    pass

  def ParseHeaderData(self):
    try:
      txt=self.header['data']
    except KeyError as e:
      #first call parseHeader()
      raise e
    self.header['data']=parsed=dict()

    pos=0
    reExpr=re.compile(r'X-Binary-([\w-]*)([\s:]*)(.*?)[\r\n]')
    while(True):
      m=reExpr.search(txt[pos:])
      if m:
        key=m.group(1)
        val=m.group(3).strip()
        #dangerous:
        #try:
        #  val=eval(val)
        #except (NameError, SyntaxError) as e:
        #  pass
        try: val=int(val)
        except (ValueError) as e:
          try: val=float(val)
          except (ValueError) as e:
            pass
        parsed[key]=val
        pos+=m.end()
      else:
        break

  def GetBinaryInfo(self):
    hdr=self.header['data']
    cmprBinSz=hdr['Size'] #X-Binary-Size
    numElem=hdr['Number-of-Elements'] #X-Binary-Number-of-Elements: 287625
    #X-Binary-Size-Fastest-Dimension: 1475
    #X-Binary-Size-Second-Dimension: 195
    dim=(hdr['Size-Second-Dimension'],hdr['Size-Fastest-Dimension'])
    #X-Binary-Size-Padding: 4095
    assert(numElem==np.prod(dim))
    if hdr['Element-Type'].find('32-bit')>=0:
      dtype=np.uint32
    elif hdr['Element-Type'].find('16-bit')>=0:
      dtype=np.uint16
    else:
      raise TypeError(hdr['Element-Type'])
    return cmprBinSz,dim,dtype

  def GetBinary(self):
    data=self.data
    cmprBinSz,dim,dtype=self.GetBinaryInfo()
    pos=self.cbfHeaderEnd+len(CbfParser.cbfHeaderMarkEnd)
    inBuf=data[pos:pos+cmprBinSz]

    outBuf=np.ndarray(dim,dtype=dtype,order='C')

    outPtr=ct.cast(outBuf.ctypes, ct.POINTER(ct.c_uint32))
    LibDetXR()
    res=LibDetXR.DecompressCBF(inBuf, cmprBinSz, outPtr, np.prod(dim)*4,outBuf.itemsize)
    return outBuf,res

  def DebugPrint(self,mode,):
    ''' print Header in various formats
       mode=2: raw
       mode=3: parsed, raw dictionaries
       mode=4: parsed, pretty dictionaries
    '''
    R=CbfParser.RESET; B=CbfParser.BOLD
    sep=CbfParser.separator
    #print "filename: ", self.fn
    header=self.header
    if mode==2:
      print B+'header:\n'+R, header['raw'], sep
    elif mode in (3,4):
      print B+"cbfVersion:"+R, '+'+header['cbfVersion']+'+'
      print B+"dataName:"+R, header['dataName']
      print B+"header_convention:"+R, header['header_convention']
      if mode==3:
        print sep+B+'\nheader_contents:\n'+R, header['header_contents']
        print sep+B+'\ndata:\n'+R, header['data']
      else:
        self.DebugPrintDict(header,'header_contents')
        self.DebugPrintDict(header,'data')

  def DebugPrintDict(self,myDict,key):
    R=CbfParser.RESET; B=CbfParser.BOLD
    sep=CbfParser.separator
    print sep+'\n'+B+key,':'+R
    for (k,v) in myDict[key].iteritems():
      print k,'\t',v
#initialize default display mode to colored
CbfParser.SetDisplayMode(True)

if __name__ == '__main__':
  import matplotlib.pyplot as plt
  import argparse #since python 2.7
  #import getpass
  #usrName=getpass.getuser()

  def RunWithArgs():

    if len(sys.argv)>1 and sys.argv[1][:4]=='Test':
      print eval(sys.argv[1]+'()')
      return

    exampleCmd="`find /scratch/detectorData/2013-03-cbf-example-files/ -name '*.cbf'` -v3 -s2"
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description=__doc__,
                                     epilog='Example:\n'+os.path.basename(sys.argv[0])+' '+exampleCmd+'\n ')
    parser.add_argument('files',  nargs='+',  help='list of cbf-files')
    #parser.add_argument('--eAccount', help='the used eAccount name')
    parser.add_argument('-v', '--verbose', type=int, nargs='?', const=1, default=0, help='verbose level:\n  0:nothing\n  1:filenames\n  2:raw header\n  3:parsed header\n  4:parsed header pretty print')
    parser.add_argument('-s', '--show',    type=int, nargs='?', const=1, default=0, help='display type:\n  1: size and average\n  2: blocking display with matplotlib\n  2: non blocking dispay with matplotlib')
    parser.add_argument('--nocolor', action='store_true', help='no coloring in verbose mode')

    args = parser.parse_args()
    print "Testing CbfParser..."
    separator = 50*'='
    if args.verbose==2:
      hdrParse=0
      if args.show:
        args.show=0
        print 'in verbose mode 2 the decoding is not done'
    else:
      hdrParse=7
    if args.verbose:
      CbfParser.SetDisplayMode(not args.nocolor)
    for fn in args.files:
      if args.verbose==1:
        print fn
      elif args.verbose>1:
        print separator
        print CbfParser.BOLD_RED+fn+CbfParser.RESET
      cbfParser=CbfParser()
      try:
        cbfParser.Open(fn)
      except (OSError,ValueError) as e:
        traceback.print_exc(1)
        #sys.stderr.write(str(e)+' : ')
        continue
      try:
        cbfParser.ParseHeader(hdrParse)
      except BaseException as e:
        traceback.print_exc(1)
        sys.stderr.write('ParseHeader Error : '+str(e)+' : ')
        continue
      if args.verbose>1:
        cbfParser.DebugPrint(args.verbose)
      if args.show>0:
        (outBuf,res)=cbfParser.GetBinary()
      if args.show==1:  #just
        R=CbfParser.RESET; B=CbfParser.BOLD
        print B+'Binary Data:'+R,outBuf.shape,np.average(outBuf)
      if args.show==2:  #blocking display
        f=plt.figure();f.suptitle(fn)
        ih=plt.imshow(outBuf,vmax=1000,interpolation='nearest')
        plt.show()
      elif args.show==3:  #non blocking display: only works if plt.ion()
        try:
          ih.set_array(outBuf)
          ih._extent=None
          ih.set_extent(ih.get_extent())
        except UnboundLocalError as e:
          plt.ion()
          f=plt.figure();f.suptitle(fn)
          ih=plt.imshow(outBuf,vmax=1000,interpolation='nearest')
        plt.draw()
      cbfParser.Close()
    pass

  def test1():
    cbfParser=CbfParser()
    fn='/scratch/detectorData/e14428/pilatus_1/S00000-00999/S00010/e14428_1_00010_00000_00000.cbf'
    cbfParser.Open(fn)
    cbfParser.ParseHeader(0)
    cbfParser.ParseHeader(7)
    (data,res)=cbfParser.GetBinary()

    cbfWr=CbfWriter()

    print '32 bit CBF writer'
    data=np.ndarray(buffer=np.random.bytes(10*20*4),shape=(10,20), dtype=np.uint32)
    data.flags.writeable=True
    data[:]/=10000
    data[:5]/=10000;data[7:]/=100;data[9:]/=100
    print data
    cbfWr.SetData(data)
    cbfWr.Write('/scratch/detectorData/out32.cbf')

    print '16 bit CBF writer'
    data=data.view(dtype=np.uint16)[:,0::2].copy()
    print data
    cbfWr.SetData(data)
    cbfWr.Write('/scratch/detectorData/out16.cbf')
    pass
#---------- MAIN TEST CODE ----------
  logging.getLogger(__name__).setLevel(logging.WARN)
  #CODECOMPLETION ONLY WORKS ON GLOBAL VARIABLES

  #usage: ./CbfParser.py `find /scratch/Documents/detectorData/ -name '*.cbf'`

  #RunWithArgs()
  test1()
