#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2014 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
This module contains

- clamps for data sources and data sinks
- functions to run source->process->destination in a single and multiprocess environment

processing utility functions/classes

 - def SingleProc(cplInType,clpOutType,proc,args):
 - def MultiProc(cplInType,clpOutType,proc,args):
 - def MultiProcChildProc(pipe,topo,jobIdx,jobCnt,proc,args):

for an example usage check procSample
'''
import logging, sys, os, glob, stat
import multiprocessing as mproc
import setproctitle as spt
import numpy as np
import utilities as ut
import h5py, zmq
import json
import tifffile
from weakref import ref as wref
from CbfFile import CbfParser

_log=logging.getLogger(__name__)

import ctypes as ct
class SharedObj(ct.Structure):
  _fields_ = [('frmCnt', ct.c_uint64), #enum ImgType
              ('cmd', ct.c_uint64), #1=stop: aborts immidiatly 2=eos: end of stream: finish when nothing remains to do
              ('foo',ct.c_ushort),
              ('bar',ct.c_ushort)]

class Clamp():
  '''Clamp base class'''
  def Close(self):
    pass
  @staticmethod
  def CombineCaps(srcCaps,dstCaps,mpMsk):
    #combine the source and destination caps and filters the mulltiprocessing caps
    caps=(srcCaps|dstCaps)
    caps=(caps&~3)|(caps&mpMsk)
    return caps
  @staticmethod
  def ShowCaps(caps):
    m=('CP','  ','CP','  ','CP')
    if caps&1:
      s0=[
's---',
'    ',
's---',
'    ',
's---']
    else:
      s0=[
'      --',
'     /  ',
's---MP--',
'     \\  ',
'      --']
    if caps&2:
      s1=[
'---d',
'    ',
'---d',
'    ',
'---d']
    else:
      s1=[
'--',
'  \\',
'--MP--d',
'  /',
'--']
    t=zip(s0,m,s1)
    u=map(lambda x: ''.join(x), t)
    for x in u:
      print x



  @staticmethod
  def GetCaps(param):
    '''returns the capabilities of this clamp

       - bit 0: can handle multiple simultaneous inputs  (e.g. true for: hdf5,zmq-pull,  false for: )
       - bit 1: can handle multiple simultaneous outputs (e.g. true for: zmq-push, false for:hdf5)
       - bit 2: can set the index of the input message (random access)
       - bit 3: can get the total number of input messages

       (MP=Master Process CP=Child Process)::

         if bit0:     else:             |   if bit1:         else:
         s---CP---            --CP---   |   ---CP---d        ---CP--
                             /          |                           \\
         s---CP---    s-----MP--CP---   |   ---CP---d        ---CP--MP--d
                             \\          |                           /
         s---CP---            --CP---   |   ---CP---d        ---CP--

    '''
    t=param[0]
    caps=1+4+8
    return caps

  @staticmethod
  def GetType(t):
    'converts type string to the numpy type: e.g. "uint32" -> numpy.uint32'
    if t==None:
      e=TypeError('need a supported data type')
      raise(e)
    try:
      dtype=np.dtype(t)
    except TypeError as e:
      raise(TypeError('need a supported data type. '+str(t)+' not understood'))
    return dtype

  @staticmethod
  def GetGlob(inGlob,ext):
    'returns a sorted list of files in a directory'
    #for recursive searching see:
    #http://stackoverflow.com/questions/14798220/how-can-i-search-sub-folders-using-glob-glob-module-in-python
    fnLst=[]
    if type(inGlob)==str: inGlob=(inGlob,)
    for glb in inGlob:
      glb=os.path.expanduser(glb)
      try:
        s=os.stat(glb)
        if stat.S_ISDIR(s.st_mode):
          glb=os.path.join(glb,'*.'+ext)
      except OSError: pass
      else: pass
      fnLst.extend(glob.iglob(glb))
    fnLst.sort()
    return fnLst

class ClampIn(Clamp):
  '''generic input clamp that supports various types as:
    randon, incremental, raw, tiff, cbf, hdf5 and zmq sources'''
  __types__='''
  Random values:
    --src rand DATATYPE   X    Y Z
    --src rand uint16  2560 2160 3
  Incremental values:
    --src inc DATATYPE   X    Y Z
    --src inc uint16  2560 2160 6
  Raw data from files .raw in a directory:
    --src raw DATATYPE   X    Y PATH
    --src raw uint16  2560 2160 /scratch/detectorData/PCO/
  TIFF data from files .tif[f] in a directory:
    --src tif PATH
    --src tif /scratch/detectorData/PCO/
  CBF data from files .cbf in a directory:
    --src cbf PATH
    --src cbf /scratch/detectorData/e14472/pilatus_1/S00000-00999/S00033/
  HDF file from a given object:
    --src hdf FILE ELEM
    --src hdf myfile.hdf5 entry/mygrp/dataset
  ZMQ source stream (now only uncompressed 'chunk'-type is supported)
    --src zmq key:value key:value (keys are addr, qs, if, timeout)
    --src zmq addr:tcp://localhost:8080 qs::4 if:PULL
'''
  'types doc'
  #in __init__ the function GetData is set
  #GetData will return the next data
  #self.idx will contain the current data index
  @staticmethod
  def GetCaps(param):
    t=param[0]
    caps=1+4+8
    if t in ('zmq','zmqMsg'):
      zmqParam=param[1]
      try:
        ifType=zmqParam['ifType']
      except KeyError:
        zmqParam['ifType']=ifType=zmq.PULL
      else:
        if type(ifType)==str:
          zmqParam['ifType']=ifType=ut.ZMQ.ifDict[ifType]
      if ifType==zmq.PULL and zmqParam['addr'].find('*')<0:
        caps=1
      else:
        caps=0
    return caps

  def SetIdx(self,idx):
    assert(self.GetData!=self.GetZmq)
    self.idx=idx-1

  def GetInfo(self):
    #if self.GetData.im_func in map(lambda x: ClampIn.__dict__[x],('GetTif','GetCbf'))
    #if self.GetData.im_func in map(lambda x: x.im_func,(ClampIn.GetTif,ClampIn.GetCbf))
    if self.GetData in(self.GetTif,self.GetCbf):
      fnLst=self.param
      return {'cnt':len(fnLst)}
    elif self.GetData==self.GetHdf:
      ds=self.param[1]
      sh=ds.shape
      return {'cnt':sh[0],'shape':sh[1:]}
    else:
      return {}

  def __init__(self, args, job=(0,1), verbose=0):
    self._job=job #index of job count of job
    self._verbose=verbose
    src=args.src
    try: self.shObj=shObj=args.shObj
    except AttributeError: pass
    t=src[0]
    func='Get'+t[0].upper()+t[1:]
    self.GetData=getattr(self,func)

    if t in ('rand','inc'):
      dtype=Clamp.GetType(src[1])
      shape=tuple(map(int,src[2:]))
      self.param=(shape,dtype)
    elif t=='raw':
      dtype=Clamp.GetType(src[1])
      shape=tuple(map(int,src[2:4]))
      path=src[4:]
      fnLst=Clamp.GetGlob(path,'raw')
      self.param=(shape,dtype,fnLst)
    elif t in ('tif', 'cbf'):
      path = src[1:]
      fnLst = Clamp.GetGlob(path, t)
      self.param = fnLst
    elif t=='oldEigerH5': # --src oldEigerH5 PATH ELEM
      path = src[1:2]
      elem = src[2]
      fnLst = Clamp.GetGlob(path, 'h5')
      fid = h5py.h5f.open(fnLst[0],flags=h5py.h5f.ACC_RDONLY)
      dsid=h5py.h5d.open(fid,elem)
      ds=h5py.Dataset(dsid)
      shape=ds.shape
      fid.close()
      self.param = (src[2],shape,fnLst)
    elif t=='hdf':
      fnHDF=src[1]
      path=src[2]
      fid = h5py.h5f.open(fnHDF,flags=h5py.h5f.ACC_RDONLY)
      dsid=h5py.h5d.open(fid,path)
      ds=h5py.Dataset(dsid)
      self.param=(fid,ds)
    elif t in('zmq','zmqMsg'): # contains raw header, and raw data
      args=src[1]
      try:
        self.numFrm=args['numFrm']
      except KeyError:
        self.numFrm=sys.maxint
      try:
        shObj
      except NameError:
        self.frmCnt=0
      self.param=ut.ZMQ.Socket(args,zmq.PULL,4)
    else:
      sys.stderr.write('unknown source: --src '+str(src))
    self.idx=job[0]-job[1]
    return

  def Close(self):
    try:
      if (type(self.param)==zmq.sugar.socket.Socket):
        sock=self.param
        #sock.disconnect(args.addr) #the socket sometimes crash.. but this also does not help :-(
        sock.close()                 #the socket sometimes crash.. but this also does not help :-(
    except AttributeError:
      pass

  #-------------- data sources virtual-functions --------------
  def GetRand(self):
    '''generates an random image and returns the array'''
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    shape,dtype=self.param
    try:
      if idx>=shape[2]:
        return None
    except IndexError as e:
      pass
    numByte=dtype.itemsize*np.prod(shape[:2])
    arr=np.random.bytes(numByte)
    arr=np.frombuffer(arr,dtype=dtype)
    arr.resize(tuple(reversed(shape[:2])))
    return arr

  def GetInc(self):
    '''generates an incremental image and returns the array'''
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    shape,dtype=self.param
    try:
      if idx>=shape[2]:
        return None
    except IndexError as e:
      pass
    numElem=np.prod(shape[:2])
    start=numElem*idx
    arr=np.arange(start,start+numElem,dtype=dtype)
    arr.resize(tuple(reversed(shape[:2])))
    return arr

  def GetRaw(self):
    '''reads all .raw file in a directory and returns the array'''
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    shape,dtype,fnLst=self.param
    if idx>=len(fnLst):
      return None
    f = open(fnLst[idx], 'r')
    arr = np.fromfile(f, dtype=dtype)
    f.close()
    arr.resize(tuple(reversed(shape)))
    if self._verbose&0x200:
      print idx,arr.shape,os.path.basename(fnLst[idx])
    return arr

  def GetTif(self):
    '''reads all .tif file in a directory and stores the data in the imgList'''
    #examples at http://code.google.com/p/pylibtiff/
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    fnLst=self.param
    if idx>=len(fnLst):
      return None
    #try:
    #  TIFF
    #except NameError:
    arr=tifffile.imread(fnLst[idx])
    #else:
    #  tif = TIFF.open(os.path.join(path,fnLst[idx]), mode='r')
    #  arr = tif.read_image()
    #  if tif.IsByteSwapped():
    #    arr.byteswap(True)
    if self._verbose&0x200:
      print idx,arr.shape,os.path.basename(fnLst[idx])
    return arr

  def GetCbf(self):
    '''reads all .cbf file in a directory and stores the data in the imgList'''
    #examples at http://code.google.com/p/pylibtiff/
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    fnLst=self.param
    if idx>=len(fnLst):
      return None
    cbfParser=CbfParser()
    cbfParser.Open(fnLst[idx])
    cbfParser.ParseHeader(1)
    (arr,res)=cbfParser.GetBinary()
    cbfParser.Close()
    if self._verbose&0x200:
      print idx,arr.shape,os.path.basename(fnLst[idx])
    return arr

  def GetOldEigerH5(self):
    (jIdx, jCnt) = self._job
    self.idx += jCnt
    idx = self.idx
    path,shape,fnLst = self.param
    lenFnLst=len(fnLst)
    iFn,iAr=divmod(idx,shape[0])
    if iFn >= lenFnLst:
      return None
    fnHDF=fnLst[iFn]
    fid = h5py.h5f.open(fnHDF, flags=h5py.h5f.ACC_RDONLY)
    dsid = h5py.h5d.open(fid, path)
    ds = h5py.Dataset(dsid)
    arr = ds[iAr,:]
    fid.close()
    if self._verbose & 0x200:
      print idx, arr.shape, os.path.basename(fnHDF), iFn, iAr
    return arr

  def GetHdf(self):
    '''reads all slices of an image stack in a hdf5 file'''
    #examples at http://code.google.com/p/pylibtiff/
    (jIdx,jCnt)=self._job
    self.idx+=jCnt
    idx=self.idx
    ds=self.param[1]
    if idx>=ds.shape[0]:
      return None
    arr=ds[idx,:]
    if self._verbose&0x200:
      print idx,arr.shape,ds.shape
    return arr

  def GetZmqMsg(self):
    sock=self.param
    numFrm=self.numFrm
    #print type(self.frmCnt)
    n=0
    while n<=self.retry:
      try:    frmCnt=self.shObj.frmCnt
      except: frmCnt=self.frmCnt
      print '>>>>>'+str(numFrm)
      try:
        if self.shObj.cmd==2: # got eos command
          if not sock.poll(0):
            break
      except AttributeError:
        pass
      if frmCnt>=numFrm:
        _log.info('Job:%d:end'%self._job[0])
        return None
      try:
        try:
          self.idx=header=ut.ConvUtf8(sock.recv_json())
        except ValueError as e: #simplejson.decoder.JSONDecodeError as e:
          _log.warning(str(e))
          return None
        if self._verbose&0x200:
          print header
        try: cmd=header['cmd']
        except KeyError as e: pass
        else:
          if cmd=='stop':
            _log.info('cmd=="stop"')
            try:
              with self.shObj.get_lock():
                self.shObj.cmd=2
            except: pass
            return None
        if sock.getsockopt(zmq.RCVMORE):
          content= sock.recv(copy=False)
          try:
            with self.frmCnt.get_lock():
              self.frmCnt.value+=1
          except: self.frmCnt+=1
          #print '>>>>>'+str(self.frmCnt)
          return content.bytes
        return None
      except zmq.error.Again as e:
        n+=1
        if n==self.retry:
          _log.warning('Job:%d:zmq timeout after %d ms'%(self._job[0],n*sock.RCVTIMEO))
        else:
          log=logging.root.handlers[0].stream
          log.write('-');log.flush()#log.write('-'+str(frmCnt)+'.'+str(numFrm))
    return None

  def GetZmq(self):
    sock=self.param
    numFrm=self.numFrm
    #print type(self.frmCnt)
    try:
      shObj=self.shObj
    except:
      frmCnt=self.frmCnt
    else:
      frmCnt=shObj.frmCnt
      while not sock.poll(100):
          if self._verbose&0x200:
            _log.info('shObj.cmd=%d'%shObj.cmd)
          if shObj.cmd>0: # got eos command and nothing left in the queue
            return None
    #_log.info(str(frmCnt))
    if frmCnt>=numFrm:
      _log.info('Job:%d:end'%self._job[0])
      return None
    try:
      try:
        self.header=header=ut.ConvUtf8(sock.recv_json())
      except ValueError as e: #simplejson.decoder.JSONDecodeError as e:
        _log.warning(str(e))
        return None
      if self._verbose&0x200:
        _log.info(str(header))
      try: cmd=header['cmd']
      except KeyError as e: pass
      else:
        if cmd=='stop':
          _log.info('cmd=="stop"')
          try:
            with shObj.get_lock():
              shObj.cmd=1
          except NameError: pass
          return None
      if sock.getsockopt(zmq.RCVMORE): #@UndefinedVariable
        content= sock.recv(copy=False)
        if header['htype'].startswith('chunk'):
          try:
            idx=int(header['frame'])#-self.frmIdxOfs #take the frame number as index in the array
          except KeyError as e:
            _log.warning('header missing "frame" entry. Data ignored!')
            return None
          try:
            swap=(header['endianess']=='big')
          except KeyError:
            swap=False
          if header.has_key('cmpr'): #compressed chunk
            _log.warning('compressed chunk as input not supported!')
            return None
          arr=np.ndarray(shape=tuple(reversed(header['shape'])),dtype=np.dtype(header['type']), buffer=content.bytes)
          if swap:#swap indianness... could be slow but works
            arr=arr.byteswap()
          arr.flags.writeable=True #make data writable even if it is read only
          self.idx=idx
          try:
            with shObj.get_lock():
              shObj.frmCnt+=1
              #_log.info(str(frmCnt)+' '+str(idx))
          except NameError: self.frmCnt+=1
          return arr
        else:
          _log.warning('htype not supported!')
    except zmq.error.Again as e:
      _log.warning('Job:%d:zmq timeout after %d ms'%(self._job[0],sock.RCVTIMEO))
    return None

class ClampOut(Clamp):
  '''base output clamp'''
  @staticmethod
  def GetCaps(param):
    return 0
  def __init__(self,**kwargs):
    for k,v in kwargs.iteritems():
      setattr(self,k,v)

  #SetData will 'print,write,send' the data to its output'
  def SetData(self,idx,data):
    _log.info('%d:'%idx+str(data)) #virtual function to overload

  def ProcRPC(self,id,cmd,*args,**kwargs):
    try: proc=self.wrefProc()
    except AttributeError:
      #pipe=self.wrefJobs[id][1]
      #pipe.send(('rpc',cmd,args,kwargs))
      #print 'ClpProcRPC',pipe.recv()
      _log.warning('PipeCom not supported:ClpProcRPC:'+cmd+str(args)+str(kwargs))
    else:
      return getattr(proc,cmd)(*args,**kwargs)
    pass

  def ClpSrcRPC(self,id,cmd,*args,**kwargs):
    '''remote procedure call to the source clamp
       use this function wisely, because it can communicate between processes'''
    try: clpSrc=self.wrefSrc()
    except AttributeError:
      #pipe=self.wrefJobs[id][1]
      #pipe.send(('rpc',cmd,args,kwargs))
      #print 'ClpSrcRPC',pipe.recv()
      _log.warning('PipeCom not supported:ClpProcRPC:'+cmd+str(args)+str(kwargs))
    else:
      return getattr(clpSrc,cmd)(*args,**kwargs)

class ClampLst:
  'helper class ClampLst contains clamps that will send data to the outer world'
  ifDict={'PUSH':zmq.PUSH,'PUB':zmq.PUB,'REP':zmq.REP}
  def __init__(self,filt=None):
    self._filt=filt
    self._clampDict=dict()

  def AddStr(self,cntrStr):
    '''parse a clamp string and add the clamp to the clamp list.

    a clap string looks like::

      <clpName>,<clpType>,<param>[,<param>...]

    clpName are:
     - raw   decompressed input data
     - proc  processed input data
     - cmpr  compressed output data

    sometimes the clamps are the same (e.g. raw,proc is identical in zmqWrite).
    for now 'zmq' is the only supported clpType

    example::

      in,zmq,addr:tcp://localhost:8090,qs:4,if:PUSH
      cmpr,zmq,addr:tcp://localhost:8091,qs:4,if:PUSH,timeout:100
      cmpr,zmq,addr:tcp://*:8092,qs:4,if:PUB,timeout:100'''
    filt=self._filt
    lst=cntrStr.split(',')
    clpName=lst[0];clpType=lst[1]
    if filt!=None and not clpName in filt:
      return # clamp filtered out
    param={}
    for v in lst[2:]:
      k,v=v.split(':',1)
      try: v=int(v)
      except ValueError: pass
      param[k]=v
    self.Add(clpName,clpType,param)

  def Add(self,clpName,clpType,param):
    assert(clpType=='zmq') #for now the only supported type
    print param
    sock=ut.ZMQ.Socket(param,zmq.PULL,4)
    try:
      sockLst=self._clampDict[clpName]
    except KeyError:
      sockLst=self._clampDict[clpName]=list()
    sockLst.append(sock)
    return sock
#@PydevCodeAnalysisIgnore

  def Send(self,clampStr,hdr,msg):
    try:
      sockLst=self._clampDict[clampStr]
    except KeyError: pass
    else:
      ClampLst._Send(sockLst,hdr,msg)
  @staticmethod
  def _Send(sockLst,hdr,msg):
    for sock in sockLst:
      sock.send_json(hdr, zmq.SNDMORE)
      sock.send(msg)

#processing utility functions/classes
#def SingleProc(cplInType,clpOutType,proc,args):
#def MultiProc(cplInType,clpOutType,proc,args):
#def MultiProcChildProc(pipe,topo,jobIdx,jobCnt,proc,args):
#for an example usage check procSample:
#cplInType:  input clamp type
#clpOutType: output clamp type
#proc:       processing class
#args:       arguments that must contain:
#             -src:     source arguments of cplInType
#             -dst:     destination arguments of clpOutType
#             -pipe:    (optional) pipe to master process (supporting commands stop and frmCnt
#             -verbose: verbosity mask: if bit 0x100 is set, basic logging of multiprocessing is turned on
#             -verbose: verbosity mask: if bit 0x200 is set, extended logging of multiprocessing is turned on
#the verbose bits are:
#   bit6(0x100): multiprocessing


def SingleProc(clpSrcType,procType,clpDstType,args):
  '''runs sourceClamp->Processing->destinationClamp in the current process.

     - if there exist the attribute args.pipe, this pipe is used for communication
     - the attribute args.src is passed to the clpSrcType
     - the attribute args is passed to the clpDstType
     - the process object receives a weak refference to th e source

     ::

      clpSrc=clpSrcType(args.src)
      clpDst=clpDstType(args,wrefSrc=wref(clpSrc),wrefProc=wref(proc))
      proc.wrefSrc=wref(clpSrc)
  '''
  clpSrc=clpSrcType(args)
  proc=procType(args)
  #clpDst=clpDstType(args,wrefSrc=wref(clpSrc),wrefProc=wref(proc))
  clpDst=clpDstType(args)
  clpDst.wrefSrc=wref(clpSrc)
  clpDst.wrefProc=wref(proc)

  pipe=args.get('pipe'); frmCnt=0
  try:
    while True:
      if pipe and pipe.poll():
        cmd=pipe.recv()
        if cmd=='stop':
          break
        elif cmd=='frmCnt':
          pipe.send(str(frmCnt))
      data=clpSrc.GetData()
      idx=clpSrc.idx
      if data is None:
        break
      res=proc.Process(idx,data);frmCnt+=1
      clpDst.SetData(idx,res)
  except KeyboardInterrupt as e:
    _log.warn('User aborted with Ctrl-C')
  proc.Close()
  clpDst.Close()
  if pipe: pipe.send(str(frmCnt)) #send the frame numbers to the pipe

def MultiProc(cplInType,procType,clpOutType,args):
  '''runs sourceClamp->Processing->destinationClamp in a multiprocess environment.

     - the attribute args.multiproc=(topo,numproc) describes the topology and the number of created the processes.

     .. seealso:: :func:`Clamp.GetCaps`, :func:`SingleProc`
     '''
  pipeMP=args.get('pipe')
  verbose=args.verbose
  srcCaps=cplInType.GetCaps(args.src)
  dstCaps=clpOutType.GetCaps(args)
  topo=args.multiproc[0]
  assert(topo>=0 and topo<=3)

  caps=Clamp.CombineCaps(srcCaps,dstCaps,topo)
  if verbose&0x100:
    Clamp.ShowCaps(caps)
  #if caps&1:   else:             |   if caps&2:       else:
  #s---CP---            --CP---   |   ---CP---d        ---CP--
  #                    /          |                           \
  #s---CP---    s-----MP--CP---   |   ---CP---d        ---CP--MP--d
  #                    \          |                           /
  #s---CP---            --CP---   |   ---CP---d        ---CP--
  jobs = []
  if not caps&1:
    clpSrc=cplInType(args,verbose=verbose)
    #srcInfo=clpSrc.GetInfo()
  if not caps&2:
    if not caps&1:
      #clpDst=clpOutType(args,wrefSrc=wref(clpSrc),wrefProc=wref(proc))
      clpDst=clpOutType(args)
    else:
      #clpDst=clpOutType(args,wrefProc=wref(proc))
      clpDst=clpOutType(args)

  try:
    maxJobs=args.multiproc[1]
  except IndexError:
    maxJobs=mproc.cpu_count()
  assert(maxJobs>0)

  args.shObj=shObj=mproc.Value(SharedObj, 0)

  for idx in range(maxJobs):
    pipeParent,pipeChild = mproc.Pipe()
    job = mproc.Process(target=MultiProcChildProc, args=(pipeChild,caps,cplInType,procType,clpOutType,idx,maxJobs,args))
    jobs.append((job,pipeParent))
    job.start()
    if idx==0:
      assert(pipeParent.recv()==1000) #wait that the first process got the first data

  if not caps&2:
    clpDst.wrefJobs=jobs
  #idx= index of processed frame
  #itr= counter of the iteration
  itr=0;frmCnt=0;
  pt=.5 if (caps&3)==3 else 0 # polltime >0 if topology 3
  run=np.zeros(maxJobs, np.uint8)
  runCnt=maxJobs
  while runCnt>0:
    #for (job,pipe) in jobs:
    #  pipe.send(('stop',0))
    if pipeMP and pipeMP.poll():
      cmd=pipeMP.recv()
      if verbose&0x100: _log.info('receive cmd:%s'%cmd)
      if cmd=='stop':
        for (job,pipe) in jobs:
          print 'send stop to',job,pipe
          pipe.send(('stop',0))
        shObj.cmd=1 #cmd 1 is stop
      elif cmd=='frmCnt':
        frmCnt=shObj.frmCnt
        if verbose&0x100: _log.info('ack frmCnt:%s'%frmCnt)
        pipeMP.send(str(frmCnt))
      continue
    jobID=itr%maxJobs
    (job,pipe)=jobs[jobID]
    if caps&2:
      if pipe.poll(pt): #if there is a message in the pipe
        msg=pipe.recv()#get the message
        if msg==None:  #check if the job will end.
          if verbose&0x100: _log.info('wait end of process %d'%job.pid)
          run[jobID]=1;job.join();runCnt-=1;itr+=1;continue
    else:
      if itr>=maxJobs:
        if pipe.poll(pt): #if there is a message in the pipe
          msg=pipe.recv()
          if msg==None:
            if verbose&0x100: _log.info('wait end of process %d'%job.pid)
            run[jobID]=1;job.join();runCnt-=1;itr+=1; _log.info('process %d finished'%job.pid);continue
          (idx,res)=msg
          clpDst.SetData(idx, res)
    if not caps&1:
      data=clpSrc.GetData()
      pipe.send((clpSrc.idx,data))
    itr+=1

  if not caps&2:
    clpDst.Close()


def MultiProcChildProc(pipe,caps,cplInType,procType,clpOutType,jobIdx,jobCnt,args):
  verbose=args.verbose
  proc=procType(args)
  s='%d:%s->%s->%s'%(jobIdx,cplInType.__name__,proc.__class__.__name__,clpOutType.__name__)
  if verbose&0x100: _log.info('start process %d:'%os.getpid()+s)
  spt.setproctitle(sys.argv[0]+' '+s)
  if caps&1:
    clpSrc=cplInType(args,job=(jobIdx,jobCnt),verbose=verbose)
  if caps&2:
    clpDst=clpOutType(args,job=(jobIdx,jobCnt))#,pipe=pipe)

  if jobIdx==0:
    pipe.send(1000)#initialization of first message done
  try: shObj=args.shObj
  except AttributeError: _log.error('>> missing shObj'+str(args))
  while True:
    if caps&1:
      if pipe.poll():
        msg=pipe.recv()
        if verbose&0x100: _log.info('CP(%d)-rcv-cmd %s'%(jobIdx,str(msg))) #receive command from pipe
        (cmd,data)=msg
        if cmd=='stop':
          try:
            with shObj.get_lock():
              shObj.cmd=1 #cmd 1 is stop
          except: pass
          break
        continue
      else:
        data=clpSrc.GetData()
        idx=clpSrc.idx
        if verbose&0x200: _log.info('CP(%d)-get-data %d:%s'%(jobIdx,idx,str(type(data))))
    else:
      msg=pipe.recv()
      if type(msg[0])!=str:
        (idx,data)=msg
        if verbose&0x200: _log.info('CP(%d)-rcv-data %d:%s'%(jobIdx,idx,str(type(data))))
      else:
        #_log.info('CP(%d)-rcv-cmd %s'%(jobIdx,str(msg))) #receive command from pipe
        if verbose&0x100: _log.info('CP(%d)-rcv-cmd %s'%(jobIdx,len(msg))) #receive command from pipe
        #if verbose: _log.warning('unknown message &s'%str(msg))
        (cmd,data)=msg
        print '>>>>>>>>',cmd
        break
        #continue
    if data is None:
    #  shObj.cmd=2 #cmd 2 end of stream
      break
    res=proc.Process(idx,data)
    if caps&2:
      clpDst.SetData(idx,res)
      if verbose&0x200: _log.info('CP(%d)-set-data %d:%s'%(jobIdx,idx,str(type(res))))
    else:
      pipe.send((idx,res))
      if verbose&0x200: _log.info('CP(%d)-snd-data %d:%s'%(jobIdx,idx,str(type(res))))


  if caps&2:
    clpDst.Close()
  pipe.send(None)#send job terminate
  proc.Close()
  if verbose&0x100: _log.info('CP %d finished'%jobIdx)

if __name__ == '__main__':
  def RunWithArgs():
    import argparse #since python 2.7
    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=(
                '--src rand uint16 2560 2160 3',
                '--src inc uint16 2560 2160 6',
                '--src raw uint16 2560 2160 /scratch/detectorData/PCO/ -v',
                '--src tif /scratch/detectorData/TestTIFF/ -v',
                '--src cbf /scratch/detectorData/e14472/pilatus_1/S00000-00999/S00033/ -v',
                '--src hdf /scratch/detectorData/e14472/scan_00033.hdf5 /entry/data/pilatus_1 -v',
                '--src zmq addr:tcp://localhost:8080 qs:4 if:PULL',
                )

    epilog= '''\n
the verbose bits are:

 - bit8(0x100): multiprocessing+clamps
 - bit9(0x200): multiprocessing+clamps (more verbose)

Source description:\n  '''+ClampIn.__doc__+':'+ClampIn.__types__+'\nExamples:'+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__, epilog=epilog)
    parser.add_argument('--src',    nargs='+', required=True, help='data source (see below for examples)')
    parser.add_argument('-v',       dest='verbose', action='store_true',    help='verbose')

    args = parser.parse_args()
    print args.__dict__
    try:
      clamp=ClampIn(args.src,args.verbose)
    except TypeError as e:
      print e
      return
    import matplotlib as mpl,matplotlib.pyplot as plt,time
    f=plt.figure(1)
    ih=None
    while True:
      arr=clamp.GetData()
      if arr==None:
        break
      #plt.imshow(arr,norm=mpl.colors.LogNorm(1,arr.max()))
      try:
        ih.set_data(arr)
      except BaseException as e:
        ih=plt.imshow(arr)
      f.canvas.set_window_title(str(arr[1:10]))
      #plt.show(block=False)
      plt.pause(0.001)

      if not args.verbose:
        sys.stdout.write('.');sys.stdout.flush()
    pass
    plt.show(block=True)

  def Test0():
    clpIn =ClampIn(['cbf','/scratch/detectorData/e14472/pilatus_1/S00000-00999/S00033/'],True)
    clpOut=ClampOut()
    while True:
      arr=clpIn.GetData()
      if arr==None or clpIn.GetIdx()>10:
        break
      clpOut.SetData(arr)

#------------------ Main Code ----------------------------------
#--src tif /scratch/detectorData/TestTIFF -v
  #for args in exampleCmd:
  #  sys.argv=sys.argv[:1]+args.split()
  #  RunWithArgs()
  RunWithArgs()
  #Test0()