#!/usr/bin/env python

# *-----------------------------------------------------------------------*
# |                                                                       |
# |  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
# |                                                                       |
# |              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
# *-----------------------------------------------------------------------*
'''
parsing .dat-files from the cSAXS-Beamline
'''

import sys,os,time,re,mmap
import numpy as np
import traceback
import logging
_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

class DatParser():
  '''Base DatParser class'''
  def Open(self,fn):
    try:
      self.Close()
    except AttributeError as e:
      pass
    if not os.path.exists(fn):
      fn=os.path.expanduser(fn)
    self.fo=open(fn,'r')#;print 'Open',fn
    pass

  def __del__(self):
    try:
      self.Close()
    except AttributeError as e:
      pass

  def Close(self):
    self.fo.close()#;print 'Close',self.fo.name

class SPECDatParser(DatParser):
  '''parse SPEC .dat files of the cSAXS beamline e.g. specES1/dat-files/specES1_started_2013_04_26_1107.dat

     variables after parsing are:

     - outVar: list of output variables
     - scan: dictionary of the scan
     - dataL:  array of labeled scan data

     A SPEC dat file looks lie this::

       #S 10  ascan  samy -5e-05 1.99995  20 0.1
       #D Fri Apr 26 16:02:11 2013
       #T 0.1  (sec)
       #Q
       #P0 0 5.215 0 0.5 0 0.3 0 0
       #P1 0.6655 -0.1345 0.2655 0.8 -6238.6438 37.61375 0.603 0.303
       #N 20
       #L samy    Epoch  curr  diode  temp  ftp  bpm4a  ...
       #C Fri Apr 26 16:02:12 2013.  Scan aborted after 0 points.
       ...
     '''

  @staticmethod
  def MakeListUnique(lst):
    '''makes the strings in a list unique by adding a number at the end of not unique strings'''
    #makes names unique
    s=set()
    for (i,e) in enumerate(lst):
      if e in s:
        j=0
        while True:
          eNew=e+str(j)
          if not eNew in s:
            lst[i]=eNew
            s.add(eNew)
            break
          j+=1
      else:
        s.add(e)
    return


  def Index(self, mode=7):
    '''Builds/rebuild an index for the SPECDatParser file

       - mode bit 0: use existing idx file
       - mode bit 1: do indexing
       - mode bit 2: write idx file

       - mode=1: use existing idx file, no indexing and writing (ignore .dat file)
       - mode=2: ignore existing idx file, do indexing, dont write idx-file
       - mode=6: ignore existing idx file, do indexing, overwrite idx file
       - mode=7: use existing idx file, do indexing, update existing idx file
       '''
    rc=False#Refcheck must be turned of for debugging
    num=8*1024
    mm=mmap.mmap(self.fo.fileno(),0,prot=mmap.PROT_READ)

    try:
      if not (mode & 1):
        raise IOError
      needInexing=self.IndexLoad()
      idxOfsS=self.idxOfsS;ofsS=self.ofsS;idxS=len(ofsS)
      ofsO=self.ofsO;idxO=len(ofsO)
      p=max(ofsS[-1],ofsO[-1])+1
    except (IOError,IndexError) as e:
      needInexing=True
      ofsS=np.empty((num),dtype=np.uint32)
      ofsO=np.empty((num),dtype=np.uint32)
      idxOfsS=-1
      idxS=0
      idxO=0
      p=0
    numStart=idxS+idxO
    if (mode & 2) and needInexing:
      #grep -b -E "#O0 |#S\s+(\d+)" /scratch/detectorData/stxm_pinhole/specES1/dat-files/specES1_started_2012_03_14_2039.log
      iter=re.finditer("#O0 |#S\s+(\d+)", mm[p:], flags=0)
      for m in iter:
        #print mm[m.start():m.end()],m.groups()
        p=m.start()
        if m.group(1)!=None:
          idxS=int(m.group(1))
          if idxOfsS<0:
            idxOfsS=idxS
          idxS-=idxOfsS
          try:
            ofsS[idxS]=p
          except IndexError as e:
            newShape=(ofsS.shape[0]+num,)
            ofsS.resize(newShape,refcheck=rc)
            ofsS[idxS]=p
          idxS+=1
        else:
          try:
            ofsO[idxO]=p
          except IndexError as e:
            newShape=(ofsS.shape[0]+num,)
            ofsO.resize(newShape,refcheck=rc)
            ofsO[idxO]=p
          idxO+=1

    ofsS.resize((idxS,),refcheck=rc)
    ofsO.resize((idxO,),refcheck=rc)

    self.idxOfsS=idxOfsS
    self.ofsS=ofsS
    self.ofsO=ofsO

    if (mode & 4) and numStart!=idxS+idxO:
      self.IndexSave()
    return

  def IndexSave(self):
    fnIdx=self.fo.name+'.idx'
    fo=open(fnIdx,'w')
    np.savez(fo,idxOfsS=self.idxOfsS,ofsS=self.ofsS,ofsO=self.ofsO)
    fo.close()

  def IndexLoad(self):
    fnIdx=self.fo.name+'.idx'
    npzfile = np.load(fnIdx)
    self.idxOfsS=npzfile['idxOfsS']
    self.ofsS   =npzfile['ofsS']
    self.ofsO=npzfile['ofsO']
    stRaw=os.fstat(self.fo.fileno())
    stIdx=os.stat(fnIdx)
    needInexing= stRaw.st_mtime > stIdx.st_mtime
    return needInexing


  def Parse(self,scanID,verbose=0):
    if scanID<=self.idxOfsS or scanID>=self.idxOfsS+len(self.ofsS):
      return False
    pS=self.ofsS[scanID-self.idxOfsS]
    try: pS2=self.ofsS[scanID-self.idxOfsS+1]
    except IndexError as e: pS2=sys.maxint
    idx=len(self.ofsO)-1
    while self.ofsO[idx]>pS:
      idx-=1
    pO=self.ofsO[idx]
    idx=0
    while self.ofsS[idx]<pO:
      idx+=1
    pO2=self.ofsS[idx]
    if verbose&32:
      print 'Found in file %s positions O:%d-%d S:%d-%d '%(self.fo.name,pO,pO2,pS,pS2)
    self.ParseRgnOutput(pO,pO2)
    self.ParseRgnScan(pS,pS2)
    return True

  def ParseRgnOutput(self,startOfs,endOfs):
    #print 'dd if=%s skip=%d bs=1 count=%d'%(self.fo.name,startOfs,endOfs-startOfs)
    #print mm[startOfs:endOfs]
    mm=mmap.mmap(self.fo.fileno(),0,prot=mmap.PROT_READ)
    iter=re.finditer("#O\d+(.*)", mm[startOfs:endOfs], flags=0)
    outLst=[]
    for m in iter:
      #print mm[m.start():m.end()],m.groups()
      outLst+=m.group(1).split()
    SPECDatParser.MakeListUnique(outLst)
    self.outVar = outLst

  def ParseRgnScan(self,startOfs,endOfs):
    #print 'dd if=%s skip=%d bs=1 count=%d'%(self.fo.name,startOfs,endOfs-startOfs)
    #print mm[startOfs:endOfs]
    #reset old data
    try: del self.dataL
    except AttributeError: pass
    self.scan=scan=dict()
    self.meta=dict()
    mm=mmap.mmap(self.fo.fileno(),0,prot=mmap.PROT_READ)
    p=startOfs
    try:
      while p<endOfs:
        p2=mm.find('\n',p)
        if p2<0:
          p2=endOfs
        ln=mm[p:p2]
        p=p2+1
        if not ln:
          continue#ignore empty lines
        elif ln[0]=='#':
          if ln[1]=='D': #Date
            self.ParseDate(ln[3:])
          elif ln[1]=='S': #scanning command
            if scan.has_key('scan'):
              break
            self.ParseScan(ln[3:])
          elif ln[1]=='P': #positions of the ouput ports
            self.ParsePosition(ln[2:])
          elif ln[1]=='L': #name of labels
            self.ParseLabel(ln[3:])
          elif ln[1]=='O': #names of the Output port
            raise( KeyError('Output ports not allowed here'))
          elif ln[1:7]=='C meta':
            self.ParseMeta(ln[8:])
          elif ln[1] in 'FECTGQNX'  :
            _log.debug('ignore'+ln[:])#ignored
          else:
            _log.warn('!UNKNOWN'+ln[:])#ignored
            pass
        else:
          self.ParseNumList(ln)
    except AssertionError as e:#  BaseException as e:
      #print sys._getframe().f_code.co_name+' Error: "'+txt+'"',e
      sys.stderr.write(ln)
      traceback.print_exc(1)
    return

  def ParseDate(self,txt):
    t=None
    for fmt in('%a %b %d %H:%M:%S %Y','%Y-%m-%dT%H:%M:%S','%Y-%b-%dT%H:%M:%S','%Y/%b/%d %H:%M:%S'):
      try:
        e=None
        t=time.strptime(txt,fmt)
        break
      except BaseException as e:
        continue
    if e:
      traceback.print_exc(1)
      sys.stderr.write('Date conversion error')
    self.scan['date']=t

  def ParseMeta(self,txt):
    try:
      k,v=txt.split(None,1)
    except ValueError as e:
      _log.warning('Meta pasing error:'+txt)
    else:
      self.meta[k]=v

  def ParsePosition(self,txt):
    prs=txt.split()
    idx=int(prs[0])
    prs=map(float,prs[1:])
    if idx==0:
      self.scan['pos']=prs
    else:
      self.scan['pos']+=prs

  @staticmethod
  def ConvType(prs,t):
    for idx,t in enumerate(t):
      idx+=2;
      if t:
        v=t(prs[idx])
        prs[idx]=v

  def ParseScan(self,txt):
    prs=txt.split()
    prs[0]=int(prs[0])
    try:
      if prs[1]=='ascan':
        SPECDatParser.ConvType(prs,(None,float,float,int,float))
      elif prs[1]=='mesh':
        SPECDatParser.ConvType(prs,(None,float,float,int,None,float,float,int,float))
      elif prs[1]=='loopscan':
        prs[2:]=map(float,prs[2:])
      elif prs[1]=='burst_scan':
        prs[2]=int(prs[2])
        prs[3:]=map(float,prs[3:])
      elif prs[1]=='round_roi_scan_sorted':
        SPECDatParser.ConvType(prs,(None,float,float,None,float,float,float,int,float))
      elif prs[1]=='round_roi':
        SPECDatParser.ConvType(prs,(None,float,float,None,float,float,float,int,float,int,int))
      elif prs[1]=='cont_line':
        SPECDatParser.ConvType(prs,(None,float,float,int,float,float))
      elif prs[1]=='ura_mesh':
        SPECDatParser.ConvType(prs,(None,float,float,int,float,None,float,float,int,float,float))
      elif prs[1]=='round_roi_scan':
        SPECDatParser.ConvType(prs,(None,float,float,None,float,float,float,int,float))
      else:
        _log.warn('unknown scan type')
    except BaseException as e:
      _log.warn('coversion error:'+str(e))

    self.scan['scan']=prs

  def ParseLabel(self,txt):
    lbl=txt.split()
    SPECDatParser.MakeListUnique(lbl)
    self.scan['label']=lbl

  def ParseNumList(self,txt):
    prs=txt.split()
    prs=map(float,prs)
    try:
      dataL=self.dataL
    except AttributeError:
      lbl=self.scan['label'] #values are unique
      dtypeLbl = np.dtype(zip(lbl, (np.float32,)*len(lbl)))
      dataL=np.array([tuple(prs)],dtype=dtypeLbl)
    else:
      sh=dataL.shape
      sh=(sh[0]+1,)
      dataL.resize(sh,refcheck=False)
      dataL[-1]=tuple(prs)
    self.dataL=dataL

class MCSParser(DatParser):
  '''parse MCS .dat files of the cSAXS beamline e.g. mcs/S00000-00999/S00033/e14472_00033.dat

     variables after parsing are:

     - header: header of the MCS table
     - dataMCS: MCS data

     an MCS file looks lie this::

        # MCS file version 0.01
        # DateTime Fri Apr 26 16:38:19 2013
        # channels 32
        # dim1 10
        # dim2 1
        # number-of-exposures 49
        # Exposure_time 0.110000 0.110000 ...
        # end-of-header
          5499985      7275         0         0         ...
          5499986      7275         0         0         ...
          ...

     '''
  def Parse(self):
    self.header=dict()
    try: del self.dataMCS
    except AttributeError: pass
    header=self.header
    try:
      for ln in self.fo.readlines():
        if len(ln)==1:
          continue#ignore empty lines
        elif ln[0]=='#':
          key=re.search('(\S*)', ln[2:]).group(0)
          if key=='MCS'  :
            header['ver']=ln[19:-1]
          elif key=='DateTime':
            self.ParseDate(ln[11:-1])
          elif key=='channels':
            header['chan']=int(ln[11:-1])
          elif key[0:3]=='dim':
            d=int(key[3:])
            v=int(ln[3+len(key):-1])
            try:
              header['dim'][d-1]=v
            except KeyError as e:
              header['dim']=[v]
            except IndexError as e:
              header['dim'].append(v)
          elif key=='number-of-exposures':
            header['numExp']=int(ln[22:-1])
          elif key=='Exposure_time':
            header['expTime']=map(float,ln[16:-1].split())
          elif key=='end-of-header':
            _log.debug('ignore'+ln[:-1])#ignored
          else:
            _log.warn('!UNKNOWN'+ln[:-1])#ignored
            pass
        else:
          self.ParseNumList(ln)
    except AssertionError as e:#  BaseException as e:
      sys.stderr.write(ln)
      traceback.print_exc(1)
    pass

  def ParseDate(self,txt):
    header=self.header
    t=None
    for fmt in('%a %b %d %H:%M:%S %Y','%Y-%m-%dT%H:%M:%S','%Y-%b-%dT%H:%M:%S','%Y/%b/%d %H:%M:%S'):
      try:
        e=None
        t=time.strptime(txt,fmt)
        break
      except BaseException as e:
        continue
    if e:
      traceback.print_exc(1)
      sys.stderr.write('Date conversion error')
    header['date']=t

  def ParseNumList(self,txt):
    prs=txt.split()
    prs=map(np.uint32,prs)
    try:
      dataMCS=self.dataMCS
    except AttributeError:
      self.dataMCS=dataMCS=np.array(prs,dtype=np.uint32)
      #dataMCS.reshape(1,len(prs))
    else:
      sh=dataMCS.shape
      if len(sh)==1:
        sh=(2,sh[0])
      else:
        sh=(sh[0]+1,sh[1])
      dataMCS.resize(sh,refcheck=False)
      dataMCS[-1,:]=prs
    return

  def FlushData(self):
    try:
      fo=self.FlushObj
    except AttributeError as e:
      header=self.header
      arr=self.arr
      used=arr.any(axis=0)
      arr=arr[:,used]
      print 'MCSParser Flush:',self.fo.name,time.asctime(header['date']),arr.shape
      res=0
    else:
      res=fo[0](self,fo[1])
    del self.arr
    self.header=dict()
    return res

class SPECLogParser(DatParser):
  '''parse SPEC .log files of the cSAXS beamline e.g. specES1/log-files/specES1_started_2013_04_26_1107.log

     in the log files all commands are logged. A SPEC log file looks lie this::

       381.SPECES1> dscan samy 2 2 40 0.1
       Total 41 points, 4.1 seconds

       #S 5  ascan  samy 2 2  40 0.1
       #D Fri Apr 26 15:59:13 2013
       #T 0.1  (sec)

     Unfortunately not all scans represent one 'scan'.

     next line shows that a mesh will produce multiple scans::

       cont_dmesh px -0.04 0.04 40 py -0.04 0.04 40 0.04
       #S 158  cont_line px -0.0407399 0.0392601 40 0.04 0.003
       #S 159  cont_line px -0.0407399 0.0392601 40 0.04 0.003
       #S 160  cont_line px -0.0407399 0.0392601 40 0.04 0.003
       ...
     '''

  def Index(self, mode=7):
    '''Builds/rebuild an index for the SPECLogParser file

       - mode bit 0: use existing idx file
       - mode bit 1: do indexing
       - mode bit 2: write idx file

       - mode=1: use existing idx file, no indexing and writing (ignore .dat file)
       - mode=2: ignore existing idx file, do indexing, dont write idx-file
       - mode=6: ignore existing idx file, do indexing, overwrite idx file
       - mode=7: use existing idx file, do indexing, update existing idx file
       '''
    rc=False#Refcheck must be turned of for debugging
    num=8*1024
    mm=mmap.mmap(self.fo.fileno(),0,prot=mmap.PROT_READ)

    try:
      if not (mode & 1):
        raise IOError
      needInexing=self.IndexLoad()
      idxOfsC=self.idxOfsC;ofsC=self.ofsC;idxC=len(ofsC)
      idxOfsS=self.idxOfsS;ofsS=self.ofsS;idxS=len(ofsS)
      p=max(ofsC[-1],ofsS[-1])+1
    except IOError as e:
      needInexing=True
      idxOfsC=-1;ofsC=np.zeros((num),dtype=np.uint32);idxC=0
      idxOfsS=-1;ofsS=np.zeros((num),dtype=np.uint32);idxS=0
      p=0
    numStart=idxS+idxC
    if (mode & 2) and needInexing:
      #grep -b -E "\.SPECES1>|#S " /scratch/detectorData/stxm_pinhole/specES1/log-files/specES1_started_2012_03_14_2039.log
      iter=re.finditer("(\d+)\.SPECES1[^>]*>|#S\s+(\d+)", mm[p:], flags=0)
      for m in iter:
        #print mm[m.start():m.end()],m.groups()
        p=m.start()
        if m.group(1)!=None:
          idxC=int(m.group(1))
          if idxOfsC<0:
            idxOfsC=idxC
          try:
            ofsC[idxC-idxOfsC]=p
          except IndexError as e:
            newShape=(ofsC.shape[0]+num,)
            ofsC.resize(newShape,refcheck=rc)
            ofsC[idxC-idxOfsC]=p
        elif m.group(2)!=None:
          idxS=int(m.group(2))
          if idxOfsS<0:
            idxOfsS=idxS
          try:
            ofsS[idxS-idxOfsS]=p
          except IndexError as e:
            newShape=(ofsS.shape[0]+num,)
            ofsS.resize(newShape,refcheck=rc)
            ofsS[idxS-idxOfsS]=p

    ofsC.resize((idxC-idxOfsC+1,),refcheck=rc)
    ofsS.resize((idxS-idxOfsS+1,),refcheck=rc)

    self.idxOfsC=idxOfsC
    self.ofsC=ofsC
    self.idxOfsS=idxOfsS
    self.ofsS=ofsS
    if (mode & 4) and numStart!=idxC+idxS:
      self.IndexSave()
    return

  def GetCmdIdxOfScanIdx(self,scanIdx):
    if scanIdx<self.idxOfsS:
      return None
    try:
      p0=self.ofsS[scanIdx-self.idxOfsS]
    except IndexError:
      return None
    #p1=self.mm.find('\n',p0)
    #ln=self.mm[p0:p1]
    #print ln
    cmdIdx=0
    while self.ofsC[cmdIdx]<p0:
      cmdIdx+=1
    cmdIdx+=self.idxOfsC-1
    return cmdIdx

  def GetCmdInfoOfCmdIdx(self,cmdIdx,verbose=0):
    mm=mmap.mmap(self.fo.fileno(),0,prot=mmap.PROT_READ)
    p0=self.ofsC[cmdIdx-self.idxOfsC]
    p1=mm.find('\n',p0)
    p2=self.ofsC[cmdIdx-self.idxOfsC+1]
    if p0==0 or p2==0:
      _log.error('failed to determine command')
      return (None,None)
    cmdStr=mm[p0:p1]
    scanIdxMin=0
    try:
      while self.ofsS[scanIdxMin]<p0:
        scanIdxMin+=1
    except IndexError:
      _log.error('failed to determine first scanID')
    scanIdxMax=scanIdxMin
    try:
      while self.ofsS[scanIdxMax]<p2:
        scanIdxMax+=1
    except IndexError:
      _log.warning('assume it is the last command')
      #ps0=self.ofsS[scanIdxMax]
      #ps1=mm.find('\n',ps0)
      #print self.mm[ps0:ps1]
    scanIdxMin+=self.idxOfsS
    scanIdxMax+=self.idxOfsS-1
    if scanIdxMin==scanIdxMax:
      scanRng=(scanIdxMin,)
    else:
      scanRng=(scanIdxMin,scanIdxMax)
    if verbose&16:
      print 'Found in file %s position command %d-%d scans %d-%d'%(self.fo.name,p0,p1,p1,p2)
      print cmdStr
      print 'command',cmdIdx,'contains scans: '+'-'.join(['%d'%k for k in scanRng])
    return (cmdStr,scanRng)

  def IndexSave(self):
    fnIdx=self.fo.name+'.idx'
    fo=open(fnIdx,'w')
    np.savez(fo,idxOfsC=self.idxOfsC,ofsC=self.ofsC,idxOfsS=self.idxOfsS,ofsS=self.ofsS)
    fo.close()

  def IndexLoad(self):
    fnIdx=self.fo.name+'.idx'
    npzfile = np.load(fnIdx)
    self.idxOfsC=npzfile['idxOfsC']
    self.ofsC   =npzfile['ofsC']
    self.idxOfsS=npzfile['idxOfsS']
    self.ofsS   =npzfile['ofsS']
    stRaw=os.fstat(self.fo.fileno())
    stIdx=os.stat(fnIdx)
    needInexing= stRaw.st_mtime > stIdx.st_mtime
    return needInexing

class OMNIDatParser(DatParser):
  '''parse OMNI .dat files of the cSAXS beamline e.g. specES1/omni/omniscan_scan_00033.dat

     next line shows the content of such a .dat file::

      Scan 00033, Samroy 0.000000
      DataPoint TotalPoints Target_x Average_x Stdev_x Target_y Average_y Stdev_y Average_angle Stdev_angle
      0.000000 200.000000 -9.769912 -9.771141 0.010487 -9.307098 -9.331218 0.010152 -0.023206 0.000709
      1.000000 200.000000 -7.829721 -7.823615 0.007828 -8.802191 -8.817383 0.003936 -0.019572 0.000838
      2.000000 200.000000 -5.869750 -5.862041 0.007257 -7.812746 -7.823331 0.005367 -0.015726 0.000718
      ...
     '''
  def Parse(self):
    self.header=dict()
    try: del self.dataONMI
    except AttributeError: pass
    header=self.header
    try:
      ln=self.fo.readline()
      prs=ln.split(',')
      for s in prs:
        k,v=s.split()
        header[k]=v
        #print k,v

      ln=self.fo.readline()
      prs=ln.split()
      #self.dataOMNI=dataOMNI=np.array(prs,dtype=np.float32)
      self.dataOMNI=np.ndarray(shape=(0,),dtype=zip(prs, (np.float32,)*len(prs)))

      for ln in self.fo.readlines():
        self.ParseNumList(ln)
    except AssertionError as e:#  BaseException as e:
      sys.stderr.write(ln)
      traceback.print_exc(1)
    pass

  def ParseNumList(self,txt):
    prs=txt.split()
    prs=tuple(map(np.float32,prs))
    dataOMNI=self.dataOMNI
    sh=dataOMNI.shape
    sh=(sh[0]+1,)
    dataOMNI.resize(sh,refcheck=False)
    dataOMNI[-1]=prs
    return

if __name__ == '__main__':
  hdlr=logging.StreamHandler()
  fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
  hdlr.setFormatter(fmt)
  del logging.root.handlers[:]
  logging.root.addHandler(hdlr)
  logging.root.setLevel(logging.DEBUG)

  logLvls={'my.class': logging.WARN,
           __name__:   logging.INFO,
          }
  for k,v in logLvls.iteritems():
    logging.getLogger(k).setLevel(v)

  class StopWatch():
    def Start(self):
      self.ts=time.time()
    def Log(self,str=None,restart=True):
      ts=time.time()
      print '%.6f'%(ts-self.ts),str
      if restart:
        self.ts=ts

  def Test1():
    """Testing SPECDatParser, MCSParser"""
    #find specES1/ -name 'specE*.dat'
    fn='/scratch/detectorData/e14428/specES1/dat-files/specES1_started_2013_02_27_0751.dat'
    parser=SPECDatParser()
    print fn
    parser.Open(fn)
    parser.Parse()

    #find mcs/ -name 'e*.dat' | sort
    fn='/scratch/detectorData/e14428/mcs/S00000-00999/S00050/e14428_00050.dat'
    fnLst=('S00000-00999/S00035/e14428_00035.dat',
           'S00000-00999/S00036/e14428_00036.dat',
           'S00000-00999/S00037/e14428_00037.dat',
           'S00000-00999/S00038/e14428_00038.dat',
           'S00000-00999/S00039/e14428_00039.dat',
           'S00000-00999/S00040/e14428_00040.dat',
           'S00000-00999/S00041/e14428_00041.dat',
           'S00000-00999/S00042/e14428_00042.dat',
           'S00000-00999/S00043/e14428_00043.dat',
           'S00000-00999/S00044/e14428_00044.dat',
           'S00000-00999/S00045/e14428_00045.dat',
           'S00000-00999/S00046/e14428_00046.dat',
           'S00000-00999/S00047/e14428_00047.dat',
           'S00000-00999/S00048/e14428_00048.dat',
           'S00000-00999/S00049/e14428_00049.dat',
           'S00000-00999/S00050/e14428_00050.dat',
           'S00000-00999/S00051/e14428_00051.dat',
           'S00000-00999/S00052/e14428_00052.dat',
           'S00000-00999/S00053/e14428_00053.dat',
           'S00000-00999/S00054/e14428_00054.dat',
           'S00000-00999/S00055/e14428_00055.dat',
           'S00000-00999/S00056/e14428_00056.dat',)


    parser=MCSParser()
    for fn in fnLst:
      fn='/scratch/detectorData/e14428/mcs/'+fn
      #print fn
      parser.Open(fn)
      parser.Parse()
      parser.Close()
    pass

  def Test2():
    """Testing SPECDatParser Indexing"""
    #find specES1/ -name 'specE*.dat'

    fn='/scratch/detectorData/specES1_started_2013_06_29_2211.dat'
    parser=SPECDatParser()
    print fn
    parser.Open(fn)
    sw=StopWatch()
    sw.Start()
    parser.Index(7)
    print parser.ofsS,parser.ofsO
    sw.Log('Indexing')
    #parser.Parse(ofsO[1],ofsS[0])
    #parser.Parse(ofsS[400],ofsS[405],False)
    #parser.Parse()#parse all
    #parser.IndexSave(fn+'.idx',ofsS,ofsO)
    #(ofsS,ofsO)=parser.IndexLoad(fn+'.idx')

    pass

  def Test3():
    """Testing SPECLogParser Indexing"""
    #find specES1/ -name 'specE*.dat'
    fn='/scratch/detectorData/stxm_pinhole/specES1/log-files/specES1_started_2012_03_14_2039.log'
    parser=SPECLogParser()
    print fn
    parser.Open(fn)
    sw=StopWatch()
    sw.Start()
    parser.Index(7)
    print parser.idxOfsC,parser.ofsC,parser.idxOfsS,parser.ofsS
    sw.Log('Indexing')
    (cmdIdx)=parser.GetCmdIdxOfScanIdx(151)
    (cmdStr,scanIdx)=parser.GetCmdInfoOfCmdIdx(cmdIdx)
    print cmdStr
    print 'command',cmdIdx,'contains scans: ',scanIdx
    #parser.Parse(ofsS[400],ofsS[405],False)
    #parser.Parse()#parse all
    #parser.IndexSave(fn+'.idx',ofsS,ofsO)
    #(ofsS,ofsO)=parser.IndexLoad(fn+'.idx')

    pass

  def Test4():
    """Testing OMNIDatParser"""
    #find specES1/ -name 'specE*.dat'
    fn='/scratch/detectorData/cSAXS_sxdm_2014_07_omny_commissioning/specES1/omni/omniscan_scan_00033.dat'
    parser=OMNIDatParser()
    print fn
    parser.Open(fn)
    parser.Parse()
    print parser.header
    print parser.dataOMNI

    parser.Close()


  #Test1()
  #Test2()
  #Test3()
  Test4()


