#!/usr/bin/env python

# *-----------------------------------------------------------------------*
# |                                                                       |
# |  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
# |                                                                       |
# |              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
# *-----------------------------------------------------------------------*
'''
Program to convert .edf files to one hdf5 file.
'''

import sys,os,time,re,mmap
import numpy as np
import ctypes as ct
import traceback
import logging
_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')

#further helpfull links are:
#http://www.esrf.eu/computing/scientific/SAXS/SaxsHeader.html
#http://xrayutilities.sourceforge.net/_modules/xrayutilities/io/edf.html#EDFDirectory.Save2HDF5

class EdfParser:
  '''
  Parse header of .edf-Files to get header and binary part.
  '''
  @staticmethod
  def parseInt(txt):
    i='[-+]?\d+'
    res=re.search('('+i+')\s*(\S*)',txt).groups()
    try:
      if res[1]:
        res=(int(res[0]),res[1])
      else:
        res=int(res[0])
    except BaseException as e:
      print sys._getframe().f_code.co_name+' Error: "'+txt+'"',e
    return res

  _keywords = {
    #key: (pattern, [value_indeces], type)
    #NON_OPTIONAL_KEYWORDS
    'EDF_BinarySize':parseInt,
    'EDF_DataBlockID':parseInt,
    'EDF_HeaderSize':parseInt,
    'ByteOrder':None,
    'Dim_1':parseInt,
    'Dim_2':parseInt,
 }

  def Open(self,fn):
    if not os.path.exists(fn):
      fn=os.path.expanduser(fn)
    self.fn=fn
    #self.fh=fh = os.open(fn, os.O_RDONLY)
    self.fs=fs=open(fn,'rb')

    self.data=mmap.mmap(fs.fileno(), 0,  access=mmap.ACCESS_READ)
    #self.fs=os.fdopen(fh)
    self.header=dict()

  def Close(self):
    del self.data
    del self.fs #this will also close the self.fh
    #os.close(self.fh)

  def ParseHeader(self):
    """
      Parsing the header
    """
    header=self.header
    fs=self.fs

    while True:
      l =fs.readline().strip()
      if l:
        break

    if l[0]!='{':
      _log.error('Wrong file header')
      return
    reExpr=re.compile(r'(\S+)\s*=\s*(.+)')
    while True:
      l =fs.readline().strip()
      m=reExpr.match(l)
      if m:
        key=m.group(1)
        val=m.group(2).rstrip('; ')
        try:
          kw=EdfParser._keywords[key]
          if type(kw)==staticmethod:
            val=kw.__func__(val)
        except KeyError as e:
          #R=EdfParser.RESET; BR=EdfParser.BOLD_RED
          #etype, value, tb = sys.exc_info(); filename, lineno, name, line =traceback.extract_tb(tb,1)[0]
          #sys.stderr.write('File "%s", line %d, in %s\n' % (filename,lineno,name))
          #sys.stderr.write(BR+'cannot parse "%s : %s"\n'%(key,val)+R)
          pass
        header[key]=val
      elif not l:
        pass
      elif l=='}':
        break
      else:
        sys.stderr.write('illegal header line: "%s"\n'%l)
        break
    pass
    self.headerSz=fs.tell()

  def GetBinary(self):
    data=self.data
    #cmprBinSz,dim,dtype=self.GetBinaryInfo()
    hdr=self.header
    #pos=hdr['EDF_HeaderSize']
    #binSz=hdr['EDF_BinarySize']
    shape=(hdr['Dim_2'],hdr['Dim_1'])
    dt=hdr['DataType']
    bo=hdr['ByteOrder']
    pos=self.headerSz

    #find the proper datatype
    if dt=="SignedByte":
        dt=np.int8
    elif dt=="SignedShort":
        dt=np.int16
    elif dt=="SignedInteger":
        dt=np.int32
    elif dt=="SignedLong":
        dt=np.int64
    elif dt=="FloatValue":
        dt=np.float32
    elif dt=="DoubleValue":
        dt=np.double
    elif dt=="UnsignedByte":
        dt=np.uint8
    elif dt=="UnsignedShort":
        dt=np.uint16
    elif dt=="UnsignedInt":
        dt=np.uint32
    elif dt=="UnsignedLong":
        dt=np.uint64
    else:
        dt=np.double

    imgIdx=0
    imgSz=shape[0]*shape[1]*np.dtype(dt).itemsize

    pos+=imgIdx*imgSz
    inBuf=data[pos:pos+imgSz]

    #if bo=='LowByteFirst':
    #elif bo=='LowByteFirst':
    outBuf=np.ndarray(shape,dtype=dt,order='C',buffer=inBuf).byteswap()
    return outBuf,0

  def DebugPrint(self,mode,):
    '''print Header in various formats

       - mode=2: raw
       - mode=3: parsed, raw dictionaries
       - mode=4: parsed, pretty dictionaries
    '''
    sep='-'*20
    #print "filename: ", self.fn
    header=self.header
    for (k,v) in header.iteritems():
      print k,'\t',v

if __name__ == '__main__':
  import matplotlib as mpl
  mpl.use('WXAgg')
  import matplotlib.pyplot as plt
  import argparse #since python 2.7
  import utilities as ut
  from libDetXR import *
  import h5py

  class HDF5Writer():
    def __init__(self,dst,cmpr,numArr):
      assert(dst[0]=='hdf')
      fnHDF5=os.path.expanduser(dst[1])
      try:
        self.fid=fid=ut.OpenFile(fnHDF5,'w')
      except IOError as e:
        _log.error('Failed to open file: '+fnHDF5)
      res,v=LibDetXR.GetVersion()
      attr=(('creator','Converted data with '+os.path.basename(__file__)),
            ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])))
      grpName=dst[2]
      grpName,dsName=dst[2].rsplit('/',1)
      self.gid=ut.CreateGrpPath(fid,grpName)
      ut.AddAttributes(self.gid,attr)
      if cmpr:
        self.CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)
      self.dsName=dsName
      self.cmpr=cmpr
      self.numArr=numArr

    def Close(self):
      ut.CloseFile(self.fid)

    def Add(self,arr):
      try:
        dsid=self.dsid
      except AttributeError:
        dtype=arr.dtype
        shape=(self.numArr,)+arr.shape
        maxshape=(h5py.h5s.UNLIMITED,)+arr.shape
        chunk=(1,)+arr.shape
        self.dsid=dsid=ut.CreateDatasetChunked(self.gid,self.dsName,dtype,shape,maxshape,chunk,self.cmpr)
        self.idx=0
      try:
        CmprChunkBuild=self.CmprChunkBuild
      except AttributeError:
        dsid.write_chunk((self.idx,0,0),arr)
      else:
        (cmprBuf,filter)=CmprChunkBuild(arr)
        dsid.write_chunk((self.idx,0,0),cmprBuf,filters=filter)
      self.idx+=1

  def RunWithArgs():

    if len(sys.argv)>1 and sys.argv[1][:4]=='Test':
      print eval(sys.argv[1]+'()')
      return

    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=("`find /scratch/detectorData/TestEDF/ -name '*.edf'` --dst hdf /tmp/out.hdf any/data -v7",)

    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description=__doc__,
                                     epilog='Example:'+''.join(map(lambda s:cmd+s,exampleCmd))+'\n ')
    parser.add_argument('files',  nargs='+',  help='list of edf-files')
    parser.add_argument('-v', '--verbose', type=int, nargs='?', const=1, default=0, help='verbose level:\n  1:filenames\n  2:header\n  4:display with matplotlib')
    parser.add_argument('--dst', nargs='+', help='the output file(s) to generate (see below for examples)')
    parser.add_argument('--cmpr',           help='compression method (zlib[lvl],lz4,lzf,cbf)')

    args = parser.parse_args()
    print "Testing EdfParser..."
    separator = 50*'='

    if args.dst:
      h5wr=HDF5Writer(args.dst,args.cmpr,len(args.files))
    for fn in args.files:
      if args.verbose>1:
        print separator
      if args.verbose&1:
        print fn
      edfParser=EdfParser()
      try:
        edfParser.Open(fn)
      except (OSError,ValueError) as e:
        traceback.print_exc(1)
        sys.stderr.write(str(e)+' : ')
        continue
      try:
        pass
        edfParser.ParseHeader()
        (outBuf,res)=edfParser.GetBinary()
      except OSError as e:
        traceback.print_exc(1)
        sys.stderr.write('ParseHeader Error : '+str(e)+' : ')
        continue
      if args.verbose&2:
        edfParser.DebugPrint(args.verbose)
        print 'Binary Data:',outBuf.shape,np.average(outBuf)
      if args.verbose&4:
        f=plt.figure();f.suptitle(fn)
        ih=plt.imshow(outBuf,interpolation='nearest')
        plt.show()
      if args.dst:
        h5wr.Add(outBuf)
      edfParser.Close()

    if args.dst:
      h5wr.Close()
    pass

#---------- MAIN TEST CODE ----------
  logging.getLogger(__name__).setLevel(logging.WARN)
  RunWithArgs()

