#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Writes a ZeroMQ stream to an HDF5-file (or to cbf-files into the destination directory)
It connects to a ZMQ address that sends json- and binary data messages.
The supported json-htype are:
 - chunk-1.0 (for raw or compressed images)
 - pilatus-1.0 (for cbf-file transfer from pilatus detector)
'''
#usefull links:
#http://stackoverflow.com/questions/7538988/zeromq-how-to-prevent-infinite-wait
import logging, sys, os
import h5py, zmq, ctypes
import numpy as np
import utilities as ut
from Clamp import ClampLst,ClampIn
from libDetXR import LibDetXR
LibDetXR()

import json,re
import multiprocessing as mproc
_log=logging.getLogger(__name__)

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()

class ZmqWriterMsgDefault:
  '''default class that is used to handle zmq messages'''

  def __init__(self,parent):
    self.parent=parent

  def Start(self,sock):
    self.sock=sock
    print 'Start',self

  def Process(self,header):
    sock=self.sock
    htype=ut.ZmqHeaderFix(header)
    if htype.startswith('dheader'):
      self.htype_dheader(sock,header)
      return 0
    elif htype.startswith('dimage'):
      self.htype_dimage(sock,header)
    elif sock.getsockopt(zmq.RCVMORE):
      content= sock.recv(copy=False)
      self.parent.AddChunk(header,content.bytes)
    return 1

  def Stop(self):
    print 'Stop',self

  def htype_dheader(self,sock,header):
    '''handling dectris zmq headers'''
    pObj=self.parent
    verbose=pObj.args.verbose
    while sock.getsockopt(zmq.RCVMORE):
      h=ut.ConvUtf8(sock.recv_json()) #general data
      if verbose&4:
        print h
      try:
        htype=h['htype']
      except KeyError:
        pass
      else:
        if (htype.startswith('dflatfield') or
            htype.startswith('dpixelmask') or
            htype.startswith('dcountrate_table')):
          content= sock.recv(copy=False)
          if verbose&4:
            print 'content: %d bytes'%len(content)
          objPath=pObj.args.dst[2].rsplit('/',1)[0]+'/'+htype
          xy=h['shape']
          dtype = np.dtype(h['type'])
          yx=tuple(reversed(xy))
          shape=maxshape=chunk=yx
          ds=ut.CreateDatasetChunked(pObj.fid,objPath,dtype,shape,maxshape,chunk,None)
          arr=np.frombuffer(content ,dtype=np.uint8)
          ds.write_chunk((0,0),arr)
          pass

  def htype_dimage(self,sock,header):
    '''handling dectris zmq headers'''
    pObj=self.parent
    verbose=pObj.args.verbose
    while sock.getsockopt(zmq.RCVMORE):
      h=ut.ConvUtf8(sock.recv_json()) #general data
      if verbose&4:
        print h
      try:
        htype=h['htype']
      except KeyError:
        pass
      else:
        if htype.startswith('dimage_d'): #dconfig "dimage_d
          content= sock.recv(copy=False)
          if verbose&4:
            print 'content: %d bytes'%len(content)
          #fix header:
          h['htype']='chunk-1.0'
          h['frame']=header['frame']
          if h['encoding']=='lz4<':
            h['cmpr']='lz4'
            del h['encoding']
            #DecompressLZ4
            #out=np.ndarray(shape=h['shape'],dtype=np.dtype(h['type']))
            #res=LibDetXR.DecompressLZ4(ctypes.c_char_p(content.bytes), len(content), out.ctypes.data, out.nbytes)
            #self.AddChunk(h,out)
            #cmpr=ut.ChunkBuildFuncLZ4(False)
            #(cmprArr,filter)=cmpr(out)
            #self.AddChunk(h,cmprArr)
            arr=np.ndarray(shape=(len(content)+16,),dtype=np.ubyte)
            arr[16:]=content #can't avoid copy: need to but 16 bytes header in front
            #write the HDF5 Header as defined in document HDF5_LZ4.pdf (HDF5 - LZ4.xlsx)
            decSz=np.prod(h['shape'])*np.dtype(np.uint32).itemsize
            arr.data[  :8 ]=np.uint64(decSz).byteswap().data #Total decompressed size
            arr.data[ 8:12]=np.uint32(decSz).byteswap().data #Decompressed bytes per block
            arr.data[12:16]=np.uint32(len(content)).byteswap().data         #Compressed length of block 0
            pObj.AddChunk(h,arr)
          else:
            pObj.AddChunk(h,content)
        if htype.startswith('dconfig'):
          pass

class ZmqWriter:

  def LoadMsgHandler(self):
    '''set the class that process zmq messages'''
    msgHandler=self.args.msgHdlr
    if msgHandler==None:
      return ZmqWriterMsgDefault(self)

    mdl=__import__(msgHandler, globals(), locals(), [msgHandler], -1)
    cls=getattr(mdl, msgHandler)
    return cls(self)

  def Run(self,args):
    self.args=args
    clpIn=ClampIn(args)

    self.zmqClpLst=zmqClpLst=ClampLst()
    for s in args.clamp:
      zmqClpLst.AddStr(s)

    if args.dst[0]=='hdf':
      fnHDF5=os.path.expanduser(args.dst[1])
      try:
        mode=args.dst[3]
      except IndexError:
        mode='w'
      fnDir=os.path.dirname(fnHDF5)
      if not os.path.exists(fnDir):
        os.makedirs(fnDir)
      try:
        self.fid=ut.OpenFile(fnHDF5,mode)
      except IOError as e:
        _log.error('Failed to open file: '+fnHDF5)
      else:
        try:
          self.ReceiveLoopHDF(clpIn)
        except KeyboardInterrupt as e:
          _log.warn('User aborted with Ctrl-C')
        self.Close(self.fid)
    elif args.dst[0]=='file':
      try:
        dstPath=os.path.expanduser(args.dst[1])
      except IndexError: dstPath=None
      else:
        if not os.path.exists(dstPath):
          os.makedirs(dstPath)
      try:
        self.ReceiveLoopFILE(clpIn,dstPath)
      except KeyboardInterrupt as e:
        _log.warn('User aborted with Ctrl-C')
    else:
      _log.error('Unknown destination')

  def ReceiveLoopFILE(self,clpIn,dstPath):
    sock = clpIn.param
    args=self.args
    pipe=args.get('pipe')
    numFrm=args.get('numFrm')
    frmCnt=0
    while True:
      try:
        #time.sleep(.5)
        if pipe and pipe.poll():
          cmd=pipe.recv()
          if args.verbose&16:
            LogOut('received '+cmd+' from '+str(pipe))
          if cmd=='stop':
            break
          elif cmd=='frmCnt':
            LogOut('send '+str(frmCnt)+' to '+str(pipe))
            pipe.send(str(frmCnt))
        header = sock.recv_json()
        try: cmd=header['cmd']
        except KeyError as e: pass
        else:
          if cmd=='stop': break
        htype=ut.ZmqHeaderFix(header)
        if sock.getsockopt(zmq.RCVMORE): #@UndefinedVariable
          content= sock.recv(copy=False)
        try:
          if not htype.startswith('pilatus'):
            raise KeyError()
        except KeyError:
          _log.error('only htype=pilatus supported')
          continue
        try: dstPath=header['path']
        except KeyError:
          if dstPath==None:
            _log.error('destination path not configured:'+str(header))
            continue
        else:
          if not os.path.exists(dstPath):
            try:
              os.makedirs(dstPath)
            except OSError as e:
              _log.error(str(type(e).__name__)+':'+str(e)); continue
        dst=os.path.join(dstPath,header['filename'])
        try:
          f = open(dst, 'wb');f.write(content.bytes);f.close()
        except IOError as e:
          #import traceback;traceback.print_exc(1)
          _log.error(str(type(e).__name__)+':'+str(e)); continue
        frmCnt+=1
        if args.verbose&4:
          LogOut(str(frmCnt)+str(header)+'\n')
        elif args.verbose&2:
          LogOut('+')
        if numFrm and frmCnt>=numFrm:
          break
      except zmq.error.Again as e:
        if args.verbose&8:
          _log.info('zmq timeout after %d ms'%sock.RCVTIMEO)
        elif args.verbose&2:
          LogOut('-')
        if pipe==None:break
    if pipe: pipe.send(str(frmCnt-1)) #send the frame numbers to the pipe
    return

  def ReceiveLoopHDF(self,clpIn):
    sock = clpIn.param
    args=self.args
    pipe=args.get('pipe')
    numFrm=args.get('numFrm')
    frmCnt=0

    msgHandler=self.LoadMsgHandler()
    msgHandler.Start(sock)

    while True:
      try:
        if pipe and pipe.poll():
          cmd=pipe.recv()
          if args.verbose&16:
            LogOut('received '+cmd+' from '+str(pipe))
          if cmd=='stop':
            break
          elif cmd=='frmCnt':
            pipe.send(str(frmCnt))
        try:
          header=ut.ConvUtf8(sock.recv_json())
        except ValueError as e: #simplejson.decoder.JSONDecodeError as e:
          if args.verbose&2:
            LogOut('*')
          elif args.verbose&4:
            _log.warning(str(e))
          continue
        try: cmd=header['cmd']
        except KeyError as e: pass
        else:
          clampDict=self.zmqClpLst._clampDict # send cmd headers to all clamps
          for clist in clampDict.itervalues():
            for c in clist:
              c.send_json(header)
          if cmd=='stop': break
        frmCnt+=msgHandler.Process(header)
        if numFrm and frmCnt>=numFrm:
          break
      except zmq.error.Again as e:
        if args.verbose&8:
          _log.info('zmq timeout after %d ms'%sock.RCVTIMEO)
        elif args.verbose&2:
          LogOut('-')
        if pipe==None:break

    if hasattr(self,'jobs'): #receive epilog
      clampDict=self.zmqClpLst._clampDict
      runProc=self.jobs[-1][0];numProc=len(self.jobs)-1
      if numProc==runProc:
        nxtProc=self.jobs[-1][1] #all process are processing. wait for the oldest
      else:
        nxtProc=0 #not all process are processing. the oldest is the first
      for i in range(numProc):
        proc,ppipe=self.jobs[nxtProc]
        if runProc>0:
          (cmprArr,filter,idx,tag)=ppipe.recv()
          oldHdr=self.jobsHdrLst.pop(0)
          try: zcLst=clampDict['cmpr']
          except KeyError: pass
          else:
            oldHdr['cmpr']=args.cmpr
            ClampLst._Send(zcLst, oldHdr, cmprArr)
          if tag:
            dsDict=self.dsDict
            dsObj=dsDict[tag]
          else:
            dsObj=self.dsObj
          dsObj.MaxReshape(idx)
          dsObj.chkIdx[dsObj.varIdx]=dsObj.ofsIdx+idx
          dsObj.ds.write_chunk(tuple(dsObj.chkIdx),cmprArr,filters=filter)
          if args.verbose&4:
            LogOut('wrote chunk %d tag %s\n'%(idx,str(tag)))
            #LogOut('tag %s chunk %d sz%d\n'%(str(tag),idx,cmprArr.nbytes))
          elif args.verbose&2:
            LogOut('w')
        nxtProc=nxtProc+1
        if nxtProc==numProc:
          nxtProc=0
        runProc-=1
        ppipe.send(None)  #terminate process
      print self.jobs
      del self.jobs, self.jobsHdrLst

    if pipe: pipe.send(str(frmCnt)) #send the frame numbers to the pipe
    msgHandler.Stop()
    return

  def AllocDataset(self,header,data):
    args=self.args
    objPath=args.dst[2]
    #_log.info(str((objPath,args.cmpr,header)))
    htype=header['htype']
    if htype.startswith('chunk'):
      xy=header['shape']
      dtype = np.dtype(header['type'])
      yx=tuple(reversed(xy))
    elif htype.startswith('pilatus'):
      from CbfFile import CbfParser
      self.cbfParser=prs=CbfParser()
      prs.SetData(data)
      prs.ParseHeader(7)
      cmprBinSz,yx,dtype=prs.GetBinaryInfo()
    else:
      _log.error('unsupported header type')
      return None

    tag=header.get('tag')
    ofsDim=getattr(args,'ofsDim',None)
    if tag:
      try:
        objPath=objPath+'/'+args.tag2path[tag]
      except (KeyError,AttributeError):
        objPath=objPath+'/'+tag
      try:
        ofsDim=ofsDim[tag]
      except (TypeError,KeyError,AttributeError):
        pass
    try:
      ofsIdx=ofsDim[0]
      preDim=tuple(ofsDim[2::2]) #the prefix dimensions
      preDimIdx=tuple(ofsDim[1::2]) #the prefix dimensions index
    except (TypeError,AttributeError,IndexError):
      ofsIdx=0
      preDim=preDimIdx=tuple()
    shape=preDim+(512,)+yx
    maxshape=preDim+(h5py.h5s.UNLIMITED,)+yx
    chunk=(1,)*(len(preDim)+1)+yx

    cmprOut=args.cmpr
    try: cmprIn=header['cmpr']
    except KeyError as e:
      cmprStep=cmprOut #if header has no compression, set the cmprSteps to cmprOut
    else:
      if not cmprOut: #if cmprOut not set, use the header compression
        cmprOut=cmprIn
        cmprStep=None
      elif cmprOut.startswith(cmprIn):
        cmprStep=cmprOut[len(cmprIn):]
        if cmprStep!='' and cmprStep[0]=='-':
          cmprStep=cmprStep[1:]
      else:
        _log.error('unsupported compression mix '+cmprIn+'->'+str(cmprOut))
        return None

    if cmprStep: #GetChunkBuildFunc only if incoming data is uncompressed and compression set
      self.ChunkBuild=ut.GetChunkBuildFunc(cmprStep)
      if args.multiproc and not hasattr(self,'jobs'):
        maxJobs=mproc.cpu_count()
        jobs = [] #numRun, nextIdx
        for idx in range(maxJobs):
          pipeParent,pipeChild = mproc.Pipe()
          p = mproc.Process(target=ZmqWriter.MultiProcChunkBuilder, args=(self.ChunkBuild,args.cbf2idx,pipeChild))
          jobs.append((p,pipeParent))
          p.start()
        jobs.append(np.zeros((2),np.uint32)) #procRun, procNxtIdx
        self.jobs=jobs
        self.jobsHdrLst=[]

    _log.info('%s %s %s ofsIdx:%d cmpr:%s'%(objPath,dtype,shape,ofsIdx,cmprOut))
    #try to get the existing dataset, or create if fail
    try:
      ds=h5py.h5d.open(self.fid,objPath)
    except KeyError:
      ds=ut.CreateDatasetChunked(self.fid,objPath,dtype,shape,maxshape,chunk,cmprOut)

    #if header['htype'].startswith('chunk'):
    #  self.frmIdxOfs=frmIdxOfs=np.uint32(header.get('frame',0))
    #  h5py.Dataset(dsid).attrs['frmIdxOfs']=frmIdxOfs
    dsObj=ut.DsObj(ds,shape,chunk,ofsIdx,len(preDimIdx))
    dsObj.chkIdx=np.zeros(len(shape),dtype=np.uint32)
    dsObj.chkIdx[0:len(preDimIdx)]=preDimIdx # is (preDimIdx,idx,0,0) for 3d dataset
    return dsObj

  def AddChunk(self,header,data):
    args=self.args
    htype=ut.ZmqHeaderFix(header)
    tag=header.get('tag')
    clampDict=self.zmqClpLst._clampDict
    try: zcLst=clampDict['in']
    except KeyError: pass
    else: ClampLst._Send(zcLst, header, data)

    if tag:
      try:
        dsDict=self.dsDict
      except AttributeError:
        dsDict=self.dsDict=dict()
      try:
        dsObj=dsDict[tag]
      except KeyError:
        dsObj=dsDict[tag]=self.AllocDataset(header,data)
    else:
      try:
        dsObj=self.dsObj
      except AttributeError:
        dsObj=self.dsObj=self.AllocDataset(header,data)
    if not dsObj: return
    if hasattr(self,'jobs'):
      runProc=self.jobs[-1][0];nxtProc=self.jobs[-1][1];numProc=len(self.jobs)-1
      proc,pipe=self.jobs[nxtProc]
      if runProc==numProc:
        (cmprArr,filter,idx,tag)=pipe.recv()
        oldHdr=self.jobsHdrLst.pop(0)
        try: zcLst=clampDict['cmpr']
        except KeyError: pass
        else:
          oldHdr['cmpr']=args.cmpr
          ClampLst._Send(zcLst, oldHdr, cmprArr)
        if tag:
          dsObj=dsDict[tag]
        dsObj.MaxReshape(idx)
        dsObj.chkIdx[dsObj.varIdx]=dsObj.ofsIdx+idx
        dsObj.ds.write_chunk(tuple(dsObj.chkIdx),cmprArr,filters=filter)
        if args.verbose&4:
          LogOut('proc%d wrote chunk %d tag %s\n'%(nxtProc,idx,str(tag)))
          #LogOut('proc%d wrote chunk %d tag %s sz %d\n'%(nxtProc,idx,str(tag),cmprArr.nbytes))
          #LogOut('tag %s chunk %d sz%d\n'%(str(tag),idx,cmprArr.nbytes))
        elif args.verbose&2:
          LogOut('w')
        runProc-=1
      pipe.send((header,data)); runProc+=1; self.jobsHdrLst.append(header)
      if args.verbose&4:
        LogOut('proc%d processing'%nxtProc+str(header)+'\n')
      nxtProc+=1;
      if nxtProc==numProc:
        nxtProc=0
      self.jobs[-1][0]=runProc;self.jobs[-1][1]=nxtProc
    else:
      if htype.startswith('chunk'):
        try:
          idx=int(header['frame'])#-self.frmIdxOfs #take the frame number as index in the array
        except KeyError as e:
          _log.warning('header missing "frame" entry. Data ignored!')
          return
        try:
          swap=(header['endianess']=='big')
        except KeyError:
          swap=False
        if swap:#swap indianness... could be slow but works
          arr=np.ndarray(shape=header['shape'],dtype=np.dtype(header['type']), buffer=data)
          arrS=arr.byteswap()
          data=arrS.data
        if header.has_key('cmpr'): #compressed chunk
          arr=np.frombuffer(data ,dtype=np.uint8)
        else:
          arr=np.frombuffer(data ,dtype=dsObj.ds.dtype)
      elif htype.startswith('pilatus'):
        idx=ZmqWriter.FileName2FrmIdx(header['filename'],args.cbf2idx)
        prs=self.cbfParser
        prs.SetData(data)
        prs.ParseHeader(7)
        arr,res=prs.GetBinary()
        try: zcLst=clampDict['raw']
        except KeyError: pass
        else: ClampLst._Send(zcLst, {'htype': 'chunk-1.0', 'frame': idx, 'shape': arr.shape, 'type': arr.dtype.name}, arr)
      else:
        _log.error('unsupported header type. ignored!')
        return
      dsObj.MaxReshape(idx)
      dsObj.chkIdx[dsObj.varIdx]=dsObj.ofsIdx+idx
      try:
        comprFunc=self.ChunkBuild
      except AttributeError as e:
        dsObj.ds.write_chunk(tuple(dsObj.chkIdx),arr)
      else:
        (cmprArr,filter)=comprFunc(arr)
        try: zcLst=clampDict['cmpr']
        except KeyError: pass
        else: header['cmpr']=args.cmpr;ClampLst._Send(zcLst, header, cmprArr)
        dsObj.ds.write_chunk(tuple(dsObj.chkIdx),cmprArr,filters=filter)
    if args.verbose&4:
      if not hasattr(self,'jobs'):
        LogOut('tag %s chunk %d\n'%(str(tag),idx))
        #LogOut('tag %s chunk %d sz%d\n'%(str(tag),idx,cmprArr.nbytes))
    elif args.verbose&2:
      LogOut('+')
    return

  @staticmethod
  def FileName2FrmIdx(fn,cbf2idx):
    '''
    convert a CSAXs filename to the frame index in the hdf5 file
    cbf2idx: this gives the conversion from cbf-filenames to index.
         it is an array of integers: ofs wgt ofs wgt ofs wgt ...
         ofs is the value substracted. wgt is the multiplier.
    FileName2FrmIdx('e13900_1_00150_00089_00005.cbf',[140 60 0 10 0 1])
      -> (150-140)*60+89*10+1*5*1=1495
    '''
    m=re.match('e\d*_(\d*)_(\d*)_(\d*)_(\d*)', fn)
    if m==None:
      _log.error('cannot convert filename to index:'+str(fn)+' return index 0')
      return 0
    fnIDX=np.array(map(int,m.groups()),dtype=np.uint32)
    try:
      cbf2idx=np.reshape(np.array(cbf2idx,dtype=np.uint32),(-1,2))
    except TypeError as e:
      _log.error('invalid cbf2idx:'+str((fn,cbf2idx)))
    ofs=cbf2idx[:,0]
    mlt=cbf2idx[:,1]
    #re.match('e\d*_(\d*)_(\d*)_(\d*)_(\d*)', fn).groups()
    fnIDX=fnIDX[-mlt.size:]-ofs
    fnIDX=np.dot(fnIDX,mlt)
    #print fn,mlt,sub,fnIDX
    return fnIDX

  @staticmethod
  def MultiProcChunkBuilder(ChunkBuildFunc,cbf2idx,pipe):
    from CbfFile import CbfParser
    cbfParser=CbfParser()
    _log.info('start PID %d'%os.getpid())
    while True:
      rcv=pipe.recv()
      if rcv==None:
        break
      (header,data)=rcv
      if header['htype'].startswith('pilatus'):
        cbfParser.SetData(data)
        cbfParser.ParseHeader(7)
        arr,res=cbfParser.GetBinary()
        idx=ZmqWriter.FileName2FrmIdx(header['filename'],cbf2idx)
        #print os.getpid(),header['filename'],idx
      else:
        idx=header['frame']
        dtype=np.dtype(header['type'])
        try:
          swap=(header['endianess']=='big')
        except KeyError:
          swap=False
        if swap:#swap indianness... could be slow but works
          arr=np.ndarray(shape=header['shape'],dtype=dtype, buffer=data)
          arrS=arr.byteswap()
          data=arrS.data
        arr=np.frombuffer(data ,dtype=dtype)
      (cmprArr,filter)=ChunkBuildFunc(arr)
      pipe.send(((cmprArr,filter,idx,header.get('tag'))))
    _log.info('end PID %d'%os.getpid())
    return 0


  def Close(self,fid):
    try:
      self.dsObj.FitShape()
    except AttributeError:
      try:
        dsLst=self.dsDict.itervalues()
        for dsObj in dsLst:
          dsObj.FitShape()
      except  AttributeError: pass
    ut.CloseFile(fid)

def RunWithArgs(args):
  #------ prepare Arguments ------
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
            'timeout':100, #100ms timeout the reveive and check for a stop command
            'cmpr':None,
            'ifType':'PULL',
            'queueSz':4,
            'verbose':3,
            'cbf2idx':None,
            'multiproc':False,
            'clamp':list(),
            'msgHdlr':None
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    #if args.multiproc==True: args.multiproc=[3]
    _log.setLevel(logging.INFO if args.verbose&1 else logging.WARNING)
  else:
    args=ut.dotdict(args.__dict__)

  #args.src=['zmq',ut.ZMQ.lst2dict(['addr:'+args.addr,'qs:%d'%args.queueSz,'if:'+args.ifType])]
  args.src=['zmq',{'addr':args.addr,'qs':args.queueSz,'if':args.ifType}]
  if args.timeout!=None:
    args.src[1]['timeout']=args.timeout
  for k in ('addr','queueSz','ifType','timeout'):
    del args[k]
  wr=ZmqWriter()
  wr.Run(args)
  pass

if __name__ == '__main__':
  #ZmqWriter.FileName2FrmIdx('e13900_1_00150_00089_00005.cbf',[140,60,0,10,0,1])
  #ZmqWriter.FileName2FrmIdx('e13900_1_00150_00214_00000.cbf',[0,1,0,0])
  import argparse #since python 2.7

  def ParseArgs():
    #default values
    ifType='PULL'
    queueSz=4

    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--addr tcp://localhost:8080 --dst hdf /scratch/detectorData/testZMQ.hdf5 entry/data/dataset --numFrames 500 --timeout 1000 -v',
                '--addr tcp://localhost:8080 --dst file /tmp/outDir --numFrames 500 --timeout 1000 -v',)

    epilog='''
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress +=frame -=timeout
   bit2(0x04): progress 'filename'
   bit3(0x08): zmq timeout
   bit4(0x10): pipe communication
   bit5(0x20):
   bit6(0x40):
   default is 1+2=3 (no -v)
   -v is 1+4+8+16=29
   -v0 is turn all off

dst (hdf|file) [param ...]  : Destinations can be:
  hdf <hdf5File> <elementPath> [mode] : this will generate a hdf5 file with the received data.
             Supported htypes are: htype=chunk, htype=pilatus
             mode describes how to create the hdf File. Default is mode=c
              - mode=c: create hdfFile and erase an old file
              - mode=a: append if existing or create hdfFile. THE DATASET MUST NOT SET EXIST !
  file [<destDir>]    : this will copy incoming htype=pilatus files to the desired destination

ofsIdx cur4d max4d cur5d max5d ...  : this indicates how to convert the frameindex to a
                                      position in the multidimensional array

clamp idStr,host:port,queueSz,ifType : comma separated clamp description string (no spaces)
             host:port,queueSz,ifType : see same parameters of input stream.
             default is: queueSz=4,ifType=PUSH
             clamp: in    input unchanged zmq message
                    raw   decompressed input data
                    proc  processed input data
                    cmpr  compressed output data
                    sometimes the clamps are the same (e.g. raw,proc is identicat in zmqWrite)
            example:  cmpr,tcp://pc9477:8093,4,PUSH proc,*:8087,7,ACK

The program can also be started as REST server. (obsolete, replaced with RESTServer.py)\n For help type:'''+cmd+'''rest -h

Example:'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__, epilog=epilog)
    parser.add_argument('--addr',                    required=True, help='host:port of the zmq connection')
    parser.add_argument('--clamp', nargs='+', default=list(), help='list of clamp,host:port,queueSz,ifType of the zmq output connection')
    parser.add_argument('--numFrames',  dest='numFrm',  type=int, default=long(h5py.h5s.UNLIMITED)>>1, help='maximal number of frames to receive')
    parser.add_argument('--ofsDim',   default=None, help='offset and multidimensional setup: ofsIdx cur4d max4d cur5d max5d')
    parser.add_argument('--cbf2idx', nargs='+', type=int, default=None, help='in case of cbf2hdf2: array to convert filename to frame index (s.a. Function FileName2FrmIdx in code)')
    parser.add_argument('--dst',       nargs='+',      required=True, help='the output file(s) to generate (see below for examples)')
    parser.add_argument('--timeout',                   type=int,      help='receive timeout in ms')
    parser.add_argument('--cmpr',  help='prefix ''s-'' for byte shuffle, compression zlib[0-9],lz4,lzf,cbf. e.g. s-zlib5: shuffle zlib level 5 compression. this is ignored if a compressed chunk is received')
    parser.add_argument('--if',  dest='ifType', choices=ut.ZMQ.ifDict,default=ifType, help='interface type (default:'+ifType+')')
    parser.add_argument('--qs',  dest='queueSz', type=int,default=queueSz, help='queue size (default:'+str(queueSz)+')')
    parser.add_argument('--multiproc', '-m',           action='store_true', help='use multiprocessing method')
    parser.add_argument('--tag2path', default=dict(),  help='JSON dictionarytag to translate tags to hdf5 path')
    parser.add_argument('-v',          dest='verbose', type=int, nargs='?', const=29, default=3, help='verbose (optional level bitmask)')
    parser.add_argument('--logFile',                   help='save to logfile instead to console')
    parser.add_argument('--msgHdlr',                   help='user class to process the JSON headers (see zmqWriterUsrMsgHandlerSample.py)')
    args = parser.parse_args()
    ut.LogConfig(args, {__name__     : -1, 'utilities'  : -1,})
    if args.tag2path:
      #--tag2path '{"AqE":"the/Path/OF_AqE"}'
      args.tag2path=json.loads(args.tag2path, object_hook=ut.ConvUtf8)
    if args.ofsDim:
      #without tags: --ofsDim '[10, 2, 5]'
      #with    tags: --ofsDim '{"AqE": [10, 2, 5]}'
      args.ofsDim=json.loads(args.ofsDim, object_hook=ut.ConvUtf8)
    #print args
    RunWithArgs(args)

#------------------ Main Code ----------------------------------
  if len(sys.argv)>=2 and sys.argv[1]=='rest':
    from zmqWriterREST import RunRESTServer
    sys.argv[0]+=' '+sys.argv[1]
    del sys.argv[1]
    RunRESTServer()
  else:
    ParseArgs()
  pass
