#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
REST-Server that can start child zmqWriter processes
'''
import sys, os, pwd, time, re, json
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from SocketServer import ThreadingMixIn
import threading, socket
from zmqWriter import _log,RunWithArgs
from libDetXR import LibDetXR
import utilities as ut
import multiprocessing as mproc

class RESTHandler(BaseHTTPRequestHandler):
#HTTP codes: http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
# 200 OK
# 201 Created
# 400 Bad Request
# 401 Unauthorized
# 404 Not Found
# 405 Method Not Allowed

  def finish(self):
    'overwrites StreamRequestHandler.finish(self) : keep the wfile open if in the sseRequests values'
    if not self.wfile.closed:
        self.wfile.flush()
    s=self.wfile
    if not s in self.server.sseRequests.itervalues():
      self.wfile.close()
    else:
      print 'keep sse connection %s of thread %s open'%(str(s),threading.currentThread())
    self.rfile.close()

  #def log_request(self, code='-', size='-'):
  #  self.log_message('%s "%s" %s %s',threading.currentThread().getName(), self.requestline, str(code), str(size))

  def log_message(self, format, *args):
    _log.info("%s [%s] %s" % (self.address_string(), self.log_date_time_string(), format%args))

  def Format(self,title,txt):
    self.send_header('Content-type',  'text/html')
    self.end_headers()
    html='<head><title>'+title+'</title></head><body><h1>'+title+'</h1><p>'+txt+'</body>'
    self.wfile.write(html)

  @staticmethod
  def Report(pipe,code,msg):
    #the Report done in function AllocAndRunWriter
    if code<300:
      _log.info(msg)
    else:
      _log.error(msg)
    if pipe:
      pipe.send((code,msg))

  @staticmethod
  def AllocAndRunWriter(verbose,args,pipe=None):
    assert(type(args)==dict)  #args must be a dict
    default={
            'verbose':verbose,
            'pipe':pipe,
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    #switch user if the process as root:
    try:
      user=args['user']
    except KeyError as e:
      _log.info('no user to set')
    else:
      try:
        spw=pwd.getpwnam(user)
      except KeyError:
        RESTHandler.Report(pipe,401,'Unknown user '+str(user))
        return
      try:
        _log.info( 'switch to user "%s" setgid %d setuid %d'%(user,spw.pw_gid,spw.pw_uid))
        os.setgid(spw.pw_gid)
        os.setuid(spw.pw_uid)
      except OSError as e:
        RESTHandler.Report(pipe,401,'setgid, setuid failes: '+str(e))
        return
    if os.getuid()==0: #Here I must not be root any more!
      RESTHandler.Report(pipe,401,'Not allowed to be root user here')
      return
    pid=os.getpid()
    RESTHandler.Report(pipe,200,'the writer started PID %d'%pid)
    _log.info('PID '+str(pid)+' args:'+str(args))#+' pipe: '+str(pipe))
    try:
      RunWithArgs(args)
    except BaseException as e:#this is the OVER-ALL-EXCEPTION_HANDLER for child processes
      import traceback as tb
      errstr='ZmqWriter PID %d crashed (%s)'%(os.getpid(),time.asctime())
      sys.stderr.write(errstr+'\n')
      tb.print_exc(file=sys.stderr)
      _log.error(errstr)
    else:
      _log.info('ZmqWriter PID %d finished'%os.getpid())
    #if pipe:
    #  frmCnt=int(pipe.recv())
    #  inst['process']=(proc.pid,('R' if alive else 'S'),frmCnt)#+proc.proc hasProc'
    #  pipe.close() # makes reading from a pipe to create an exception

  @staticmethod
  def InstWait(inst,numFrm,timeout):
    frmCnt=0
    if inst.has_key('process'):
      proc,pipe=inst['process']
      alive=proc.is_alive()
      if alive:
        timeout=float(timeout)/1000.
        t=time.time()
        while True:
          pipe.send('frmCnt')
          if pipe.poll(timeout):
            frmCnt=int(pipe.recv())
            if frmCnt>=numFrm:
              break
          else:
            break
          if time.time()-t>timeout:
            break
          time.sleep(.1)
      else:
        try:
          frmCnt=inst['frmCnt']
        except KeyError:
          if pipe.poll(0):
            frmCnt=int(pipe.recv())
          else:
            _log.error('failed to get frmCnt')
          inst['frmCnt']=frmCnt
    return frmCnt

  @staticmethod
  def InstJSON(inst):
    'returns a JSON string of the instance'
    if inst.has_key('process'):
      instJSON=inst.copy()
      proc,pipe=instJSON.pop('process')
      alive=proc.is_alive()
      if alive:
        pipe.send('frmCnt')
        if pipe.poll(1):
          frmCnt=int(pipe.recv())
        else:
          frmCnt=0
      else:
        try:
          frmCnt=instJSON.pop('frmCnt')
        except KeyError:
          if pipe.poll(1):
            frmCnt=int(pipe.recv())
          else:
            frmCnt=0;_log.error('failed to get frmCnt')
          inst['frmCnt']=frmCnt
      instJSON['process']=(proc.pid,('R' if alive else 'S'),frmCnt)#+proc.proc hasProc'
    else:
      instJSON=inst.copy()
      try: frmCnt=instJSON.pop('frmCnt')
      except KeyError: frmCnt=None

      instJSON['process']=(None,None,frmCnt)#+proc.proc hasProc'
    return json.dumps(instJSON)

  def GetHtml(self,fileName):
    'serves the file relative to this script and sends this content as html object'
    try:
      fn=os.path.join(os.path.dirname(sys.argv[0]),fileName)
      fh=open(fn)
      html=fh.read()
      fh.close()
    except IOError:
      self.send_error(404,'Resource not found: %s' % self.path)
    else:
      self.send_response(200)
      self.send_header('Content-type',  'text/html')
      self.end_headers()
      self.wfile.write(html)

  def SendSSE(self,msg,data):
      errRqLst=[]
      fmt='event: %s\ndata: %s\n\n'
      buf = fmt%(msg,data)
      buf = hex(len(buf))[2:]+'\r\n'+str(buf)+'\r\n'
      for sseRq,client in self.server.sseRequests.iteritems():
        try:
          client.write(buf)
        except (AttributeError,socket.error) as e:
          errRqLst.append(sseRq)
      for sseRq in errRqLst:
        client=self.server.sseRequests.pop(sseRq)
        _log.warning('remove closed sse session'+str(client))
        try:
          sseRq.close()
          client.close()
        except (AttributeError,socket.error) as e:
          _log.warning(str(e))

  def do_GET(self):
    if re.match('/*$',self.path):
      self.GetHtml('zmqWriter.html')
      return
    r=re.match('/([\w\.-]+)/*$',self.path)
    if not r:
      self.GetHtml(self.path)
      return
    name=r.group(1)
    if name=='info':
      self.send_response(200)
      self.send_header('Content-type',  'application/json')
      self.end_headers()
      LibDetXR.GetVersion()
      content=dict()
      content['version']=tuple(map(int,reversed(LibDetXR.GetVersion()[1])))
      self.wfile.write(json.dumps(content))
    elif name=='sse':
      #print 'sse',self,self.wfile,self.rfile
      self.send_response(200)
      self.send_header('Content-Type', 'text/event-stream')
      self.send_header('Transfer-Encoding','chunked')
      self.end_headers()
      #register to main server so the sockets are not closed
      self.server.sseRequests[self.request]=self.wfile
    else:
      instances=self.server.instances
      if name=='all':
        content='{'
        for k,v in instances.iteritems():
          content+='"'+k+'":\n  '
          content+=RESTHandler.InstJSON(v)+',\n'
        if len(instances)>0:
          content=content[:-2]
        content+='}\n'
      else:
        try:
          obj=instances[name]
        except KeyError as e:
          self.send_error(404,'Resource not found: %s' % self.path)
          return
        content=RESTHandler.InstJSON(obj)
      self.send_response(200)
      self.send_header('Content-type',  'application/json')
      self.end_headers()
      self.wfile.write(content)
      self.SendSSE('msg', 'get '+name)

  def GetInst(self,name):
    instances=self.server.instances
    try:
      inst=instances[name]
    except KeyError as e:
      self.send_error(404,'Resource not found: %s allocate with e.g. /zmqWriter/instance1/config' % self.path)
      return None
    return inst

  def ConfigJSON(self,name,postBody):
    instances=self.server.instances
    try:
      inst=instances[name]
    except KeyError as e:
      if not postBody:
        self.send_error(404,'Resource not found: %s allocate with e.g. /zmqWriter/instance1/config' % self.path)
        return None
    else:
      try:
        proc,pipe=inst['process']
      except KeyError as e: pass
      else:
        if proc.is_alive():
          self.send_error(404,'the writer is already running: PID %d'%proc.pid)
          return None
    if postBody:
      try:
        inst=json.loads(postBody, object_hook=ut.ConvUtf8)
      except (ValueError) as e:
        self.send_error(400,'illegal json string: '+name+' : '+postBody)
        return None
      instances[name]=inst
    return inst

  def do_PUT(self):
    cmdLst=('config','runblk','run','stop','wait')
    r=re.match('/([\w\.-]+)/(\w+)/*$',self.path)
    if r==None:
      self.send_error(404,'Resource not found: %s\nexpect e.g. "/zmqWriter/instance1/config' % self.path)
      return
    else:
      name,cmd=r.groups()
      if not cmd in cmdLst:
        self.send_error(404,'Unknown command %s\nsupported:'%cmd+', '.join(cmdLst))
        return
      reserved=('info','sse')
      if name in reserved:
        self.send_error(400,'instance name "'+name+'" is reserved! ')
        return
    try:
      contentLen = int(self.headers.getheader('content-length'))
      postBody = self.rfile.read(contentLen)
    except TypeError:
      postBody=None
    if cmd=='config':
      if not self.ConfigJSON(name,postBody):
        return
      code=201; res='config instance %s done\n'%name
    else:
      if cmd=='runblk': #Not starting a new process. For Tests only.
        # !!! The server will block an no commands will be received any more !!!
        inst=self.ConfigJSON(name,postBody)
        if inst==None:
          return
        RESTHandler.AllocAndRunWriter(self.server.verbose, inst)
        code=200; res='runblk done\n'
      elif cmd=='run':
        inst=self.ConfigJSON(name,postBody)
        if inst==None:
          return
        pipeParent,pipeChild = mproc.Pipe()
        proc = mproc.Process(target=RESTHandler.AllocAndRunWriter, args=(self.server.verbose, inst, pipeChild))
        inst['process']=(proc,pipeParent)
        proc.start()
        (code,res)=pipeParent.recv()
      elif cmd=='stop':
        inst=self.GetInst(name)
        if inst==None:
          return
        try:
          proc,pipe=inst['process']
        except KeyError as e:
          self.send_error(404,'process not running: "%s"'%self.path)
          return
        if self.server.verbose&1:
          _log.info('send stop to PID %d(%s) pipe %d'%(proc.pid,'R' if proc.is_alive() else 'S', pipe.fileno()))
        pipe.send('stop')
        proc.join()
        del inst['process']
        code=200; res='stop done\n'
      elif cmd=='wait':
        try:
          postJSON=json.loads(postBody, object_hook=ut.ConvUtf8)
        except (ValueError) as e:
          self.send_error(400,'illegal json string: '+name+' : '+postBody)
          return
        inst=self.GetInst(name)
        if inst==None:
          return
        frmCnt=RESTHandler.InstWait(inst,postJSON['frmCnt'],postJSON['timeout'])
        code=200 if frmCnt>=postJSON['frmCnt'] else 408; res='frmCnt:%d\n'%frmCnt
        self.send_response(code,res)
        return
    self.send_response(code,res)
    self.SendSSE('cmd', 'UpdateInstLst')
    self.SendSSE('msg', cmd+' '+name)

  def do_DELETE(self):
    instances=self.server.instances
    r=re.match('/([\w\.-]+)/*$',self.path)
    if r==None:
      self.send_error(404,'Resource not found: %s' % self.path)
      return
    name=r.group(1)

    if name=='all':
      #for k,v in instances.iteritems(): The line below seems buggy
      for inst in instances:
        try: proc,pipe=inst['process']
        except KeyError: continue
        if self.server.verbose&1:
          _log.info('send stop to PID %d(%s) pipe %d'%(proc.pid,'R' if proc.is_alive() else 'S', pipe.fileno()))
        pipe.send('stop')
        proc.join()
      instances.clear()
      res='DELETE all instances done\n'
    else:
      try:
        inst=instances.pop(name)
      except KeyError as e:
        self.send_error(404,'Resource not found: %s' % self.path)
        return
      try: proc,pipe=inst['process']
      except KeyError: pass
      else:
        if self.server.verbose&1:
          _log.info('send stop to PID %d(%s) pipe %d'%(proc.pid,'R' if proc.is_alive() else 'S', pipe.fileno()))
        pipe.send('stop')
        proc.join()
      res='DELETE instance %s done\n'%name
    self.send_response(200,res)
    self.SendSSE('cmd', 'UpdateInstLst')
    self.SendSSE('msg', 'delete '+name)

class RESTServer(HTTPServer):
  def __init__(self, server_address, RequestHandlerClass):
    HTTPServer.__init__(self, server_address, RequestHandlerClass)
    self.sseRequests={}

  def handle_error(self, request, client_address):
    'this overwrites the function in class HTTPServer'
    etype, value, tb = sys.exc_info()
    if etype==socket.error:
      _log.error(str((threading.currentThread().getName(),value,client_address)))
    else:
      HTTPServer.handle_error(self, request, client_address)

  def shutdown_request(self, request):
    if not request in self.sseRequests.keys():
      #print 'shutdown_request',request
      HTTPServer.shutdown_request(self, request)
    else:
      print 'shutdown_request: keep sseRequests ',request

  def close_request(self, request):
    if not request in self.sseRequests.keys():
      #print 'close_request',request
      HTTPServer.close_request(self, request)
    else:
      print 'close_request: keep sseRequests ',request

class ThreadedHTTPServer(ThreadingMixIn, RESTServer):
  'Handles the http requests in separate threads'



def RunRESTServer():
  '''

Examples of http-requests:
  >>>get/delete all instances<<<
  curl -i -X GET http://localhost:8081/all
  curl -i -X DELETE http://localhost:8081/all  (TODO: IS BUGGY)

  >>>create/get/delete a instance with name 'cam1' instances<<<
  curl -i -X PUT http://localhost:8081/cam1/config -d '{"addr":"tcp://localhost:8080", "dst":["hdf","/scratch/detectorData/testZMQ.hdf5","entry/data/dataset"]}'
  curl -i -X GET http://localhost:8081/cam1
  curl -i -X DELETE http://localhost:8081/cam1

  >>>run/stop the existing instance 'cam1'<<<
  curl -i -X PUT http://localhost:8081/cam1/run (start non blocking in own process)
  curl -i -X PUT http://localhost:8081/cam1/runblk (start blocking in same process)
  curl -i -X PUT http://localhost:8081/cam1/stop (stop the writer)
  curl -i -X PUT http://localhost:8081/cam1/wait -d '{"timeout":1000, "frmCnt":40}'

  >>>it is also possible to run/runblk with configuration<<<
  curl -i -X PUT http://localhost:8081/cam1/run -d '{"addr":"tcp://localhost:8080", "dst":["hdf","/scratch/detectorData/testZMQ.hdf5","entry/data/dataset"]}'


Warning:
  - 'GET http://localhost:8081/cam1'
    If the process is running, it responds after the next frame and would timeout
    after one second. During this time the REST server is blocked.
    If the process is stopped (or did not yet start), the server responds immediately.

  - 'PUT http://localhost:8081/cam1/wait'
    This command is blocking! As long as a wait command is finished,
    the REST server does not respond to any answer!
    wait is finished when the frame count of times out is reached.

The configuration is a json dictionary with the same entries as the command line:
  numFrm dst timeout verbose cmpr ifType queueSz
  default values are:
 'timeout':100,
 'cmpr':None,
 'ifType':'PULL',
 'queueSz':4,
 'verbose':(same as server),

internet addresses:
  http://localhost:8090/ html page with all informations
  http://localhost:8090/all JSON string of all instances
  http://localhost:8090/cam1 JSON string of instance cam1
  http://localhost:8090/info JSON string with information as version of the library
  http://localhost:8090/sse server side event address

Examples:
  curl -i -X PUT http://localhost:8081/cam1/config -d '{"addr":"tcp://localhost:8080", "dst":["hdf","/scratch/detectorData/testZMQ.hdf5","entry/data/dataset"]}'
  curl -i -X PUT http://localhost:8090/cam1/config -d '{"addr":"tcp://localhost:8080", "dst":["file"], "numFrm":500, "timeout":1000, "verbose":1, "ifType":"PULL"}'
  curl -i -X PUT http://localhost:8090/cam0/config -d '{"addr":"tcp://localhost:8080", "dst":["hdf","/scratch/detectorData/testZMQ.hdf5","entry/data/dataset"], "numFrm":100,"cbf2idx":[0, 1, 0, 0],"multiproc":true,"cmpr":"s-zlib4"}'
 '''

  #good example for POST PUT and DELETE
  #http://blogs.plexibus.com/2009/01/15/rest-esting-with-curl/
  #http://curl.haxx.se/docs/httpscripting.html
  #http://www.redmine.org/projects/redmine/wiki/Rest_api_with_curl
  exampleCmd=('rest',
              'rest --port=8001 -v')
  cmd='\n  '+os.path.basename(sys.argv[0])+' '

  epilog='Examples:'+''.join(map(lambda s:cmd+s,exampleCmd))+RunRESTServer.__doc__

  port=8081
  import argparse #since python 2.7
  parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__, epilog=epilog)
  parser.add_argument('--port',                      type=int, nargs='?', default=port, help='port of the REST server (default%d)'%port)
  parser.add_argument('-v',          dest='verbose', type=int, nargs='?', const=29, default=3, help='verbose (optional level bitmask)')
  parser.add_argument('--logFile',                   help='save to logfile instead to console')

  args = parser.parse_args()
  ut.LogConfig(args, {__name__     : -1, 'utilities'  : -1,})
  try:
      server = RESTServer(('', args.port), RESTHandler) #Single Threaded Server: USE TO DEBUG
      #server = ThreadedHTTPServer(('', args.port), RESTHandler) #Multi Threaded Server: USE FOR RELEASE
      server.verbose=args.verbose
      server.instances={} #instances REST-objects to the user
      if args.verbose&1:
        _log.info('started REST server on port %d...'%args.port)
      server.serve_forever()
  except KeyboardInterrupt:
      if server.verbose&1:
        _log.info('^C received, shutting down server')
      server.socket.close()

  #kill remaining child processes
  for inst in server.instances.itervalues():
    try:
      proc,pipe=inst['process']
    except KeyError: pass
    else:
      if server.verbose&1:
        _log.info('send stop to PID %d(%s) pipe %d'%(proc.pid,'R' if proc.is_alive() else 'S', pipe.fileno()))
      pipe.send('stop')
      proc.join()


if __name__ == '__main__':
  RunRESTServer()
