#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2014 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
converts .hdf5 files from dectris to .cbf files skipping and summing a desired amount of images.
'''

import sys,os,time,glob,stat,json,re
import numpy as np
import traceback
import logging
import h5py
import utilities as ut
from libDetXR import LibDetXR
import multiprocessing as mproc

_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

from CbfFile import *

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()


class DectrisHDF2Cbf():
  def __init__(self,args):
    self.args=args
  def Run(self):
    args=self.args
    verb=args.verbose
    try: os.makedirs(args.outDir)
    except OSError: pass
    fnTpl=os.path.join(args.outDir,'out_%.6d.cbf')

    if args.multiproc:
      jobs = []
      maxJobs=args.multiproc
      if maxJobs<0:
        maxJobs=mproc.cpu_count()
      assert(maxJobs>0)
      for idx in range(maxJobs):
        p = mproc.Process(target=DectrisHDF2Cbf.ExportCBF, args=(self,fnTpl,idx,maxJobs))
        jobs.append((p,None))
        p.start()
      _log.info('waiting child processed...')
      for (p,pipeParent) in jobs:
        #(idx,res)=pipeParent.recv()
        p.join()
      _log.info('finished')
      pass
    else:
      self.ExportCBF(fnTpl,0,1)

  def OpenHDF(self):
    args=self.args
    verb=args.verbose

    fapl = h5py.h5p.create(h5py.h5p.FILE_ACCESS)
    #changing cache size makes no speed changes
    #_log.info('cache settings (old):'+str(fapl.get_cache()))
    #fapl.set_cache(0,2**15-1,2**23,0.75) #set big chunk cache
    #_log.info('cache settings (new):'+str(fapl.get_cache()))
    fid = h5py.h5f.open(args.hdfFile, flags=h5py.h5f.ACC_RDONLY)

    if not args.elem:
      args.elem='entry'
    oid=h5py.h5o.open(fid,args.elem)

    if type(oid)==h5py.h5d.DatasetID:
      dsLst=np.ndarray(1,dtype=object)
      dsid=oid
      dsLst[0]=h5py.Dataset(dsid)
      numFrm=dsFrmPerEntry=dsLst[0].shape[0]
    else:
      gid=oid
      num=gid.get_num_objs()
      dsLst=np.ndarray(num+1,dtype=object)
      rng=np.array([2**15-1,0],dtype=np.uint16)
      for oIdx in range(gid.get_num_objs()):
        hidStr=gid.get_objname_by_idx(oIdx)
        m=re.match('data_(\d*)',hidStr)
        if m:
          try:
            dIdx=int(m.group(1))
            dsid=h5py.h5o.open(gid,hidStr)
            ds=h5py.Dataset(dsid)
            dsLst[dIdx]=ds
            #print dIdx
            if dIdx<rng[0]: rng[0]=dIdx
            if dIdx>rng[1]: rng[1]=dIdx
          except KeyError as e:
            print str(e)
      dsLst=dsLst[rng[0]:rng[1]+1]
      dsFrmPerEntry=dsLst[0].shape[0]
      numFrm=(dsLst.shape[0]-1)*dsFrmPerEntry+dsLst[-1].shape[0]

    self.dsLst=dsLst
    self.numFrm=numFrm
    self.dsFrmPerEntry=dsFrmPerEntry
    return

  def ExportCBF(self,fnTpl,idxJob,numJobs):
    self.OpenHDF()
    args=self.args
    verb=args.verbose
    dsLst=self.dsLst
    dsFrmPerEntry=self.dsFrmPerEntry

    cbfWr=CbfWriter()

    try:
      try:
        (nOfs,nSum0,nSkip,nSum1,div)=args.proc['osssd']
      except KeyError:
        nOfs=0;(nSum0,nSkip,nSum1,div)=args.proc['sssd']
    except KeyError:
      num=self.numFrm
      numPerJob,remJob=divmod(num,numJobs)
      if idxJob<remJob:
        numPerJob+=1
        start=numPerJob*idxJob
      else:
        start=numPerJob*idxJob+remJob
      if verb&4:
        _log.info('idxJob:%d, numJobs:%d, numPerJob:%d, start:%d, num:%d'%(idxJob ,numJobs, numPerJob, start, num))

      for frmIdx in range(start,start+numPerJob):
        (dsLstIdx,dsIdx)=divmod(frmIdx,dsFrmPerEntry)
        ds=dsLst[dsLstIdx]
        data=ds[dsIdx,...]
        data=np.array(data,dtype=np.uint32)
        cbfWr.SetData(data)
        fn=fnTpl%(frmIdx+1)
        cbfWr.Write(fn)
        if verb&4:
          _log.info(fn+' frm %d'%frmIdx)
        elif verb & 2:
          LogOut('.')
    else: #(nSum0,nSkip,nSum1,div)=args.proc['sssd']
      nBlk=(nSum0+nSkip)*nSum1
      num=(self.numFrm+nSkip-nOfs)/nBlk
      numPerJob,remJob=divmod(num,numJobs)
      if idxJob<remJob:
        numPerJob+=1
        start=numPerJob*idxJob
      else:
        start=numPerJob*idxJob+remJob
      step=(nSum0+nSkip)*nSum1
      sl=(
          np.arange(0,nSum0,dtype=np.uint32).reshape(1,-1)+
          np.arange(0,nSum1,dtype=np.uint32).reshape(-1,1)*(nSum0+nSkip)
         ).ravel()
      if verb&4:
        _log.info('idxJob:%d, numJobs:%d, numPerJob:%d, start:%d, num:%d'%(idxJob ,numJobs, numPerJob, start, num))
      if verb&8 and idxJob==0:
        _log.info('step %d slicing %s'%(step,str(sl)))

      #CheckSlicingTime(dsLst[0]);return

      #dtypeAcc=np.float64
      dtypeAcc=np.uint32
      if dtypeAcc==np.uint32:
        div=np.uint32(div)

      for idx in range(start,start+numPerJob):
        frmIdx=nOfs+idx*step
        (dsLstIdx0,dsIdx0)=divmod(frmIdx,dsFrmPerEntry)
        (dsLstIdx1,dsIdx1)=divmod(frmIdx+step-nSkip-1,dsFrmPerEntry)
        if dsLstIdx0==dsLstIdx1:
          ds=dsLst[dsLstIdx0]
          if nSum0==1:
            data=ds[dsIdx0:dsIdx1:(nSkip+1),...]
            data=np.sum(data, 0, dtypeAcc)
          elif nSum1==1:
            data=ds[dsIdx0:dsIdx0+nSum0:1,...]
            data=np.sum(data, 0, dtypeAcc)
          else:
            #correct but very slow !!!
            #data=ds[sl+dsIdx0,...]

            #faster solution: and still the fastest!
            data=ds[dsIdx0:dsIdx0+sl[-1]+1,...]
            data=data[sl,...]
            data=np.sum(data, 0, dtypeAcc)

            #other faster solution
            #for i in range(nSum1):
            #  sl1=sl[i*nSum0:(i+1)*nSum0]
            #  v=ds[sl1,...]
            #  print v.shape,sl1
            #  if i==0:
            #    data=np.sum(v, 0, dtypeAcc)
            #  else:
            #    data+=np.sum(v, 0, dtypeAcc)

        else:
          if verb&8:
            _log.info('mix datasets'+str((dsLstIdx0,dsIdx0,dsLstIdx1,dsIdx1)))
          assert(dsLstIdx1==dsLstIdx0+1) #only allow to mix 2 sets
          ds0=dsLst[dsLstIdx0]
          ds1=dsLst[dsLstIdx1]
          if nSum0==1:
            data0=ds0[dsIdx0::(nSkip+1),...]
            data1=ds1[:dsIdx1:(nSkip+1),...]
          elif nSum1==1:
            try:
              data0=ds0[dsIdx0:,...]
              data1=ds1[:dsIdx1+1,...]
            except ValueError as e:
              print ds0,dsIdx0,nSum1
          else:
            slMix=sl+dsIdx0
            n0=np.nonzero(slMix<dsFrmPerEntry)[0].size
            data0=ds0[dsIdx0:dsIdx0+sl[n0-1]+1,...]
            data0=data0[sl[:n0],...]
            data1=ds1[:dsIdx1+1,...]
            data1=data1[dsIdx0-dsFrmPerEntry+sl[n0:],...]
          data0=np.sum(data0, 0, dtypeAcc)
          data1=np.sum(data1, 0, dtypeAcc)
          data=data0+data1

        #data[data==65535*numSum]=65535
        #if div!=1:
        #  data/=div
        if dtypeAcc!=np.uint32:
          data/=div
          data=np.uint32(data.round())
        else:
          data/=div

        cbfWr.SetData(data)
        fn=fnTpl%(idx+1)
        cbfWr.Write(fn)
        if verb&4:
          _log.info(fn+' frm(%d-%d)'%(frmIdx,frmIdx+step-1))
        elif verb&2:
          LogOut('.')
    if verb&2:
      LogOut('\n')
    return

if __name__ == '__main__':
  def LogConfig(args):
    logging.basicConfig()
    hdlr=logging.StreamHandler()
    fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
    #fmt = logging.Formatter('%(levelname)s:%(message)s')
    hdlr.setFormatter(fmt)
    del logging.root.handlers[:]
    logging.root.addHandler(hdlr)
    logging.root.setLevel(logging.DEBUG)
    if args.verbose&1:
      lvl=logging.INFO
    else:
      lvl=logging.WARNING

    logLvls={'CbfFile'    : logging.WARN,
             'DatParser'  : logging.WARN,
             'utilities'  : logging.WARN,
             'libDetXR'   : logging.WARN,
             __name__     : logging.INFO,
            }
    for k,v in logLvls.iteritems():
      logging.getLogger(k).setLevel(v)
    pass

  import argparse #since python 2.7
  def RunWithArgs():

    if len(sys.argv)>1 and sys.argv[1][:4]=='Test':
      print eval(sys.argv[1]+'()')
      return

    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_data_000001.h5 --elem entry/data -o /scratch/detectorData/tmpOut',
                '-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_data_000001.h5 --elem entry/data -o /scratch/detectorData/tmpOut -p \'{"sssd":[5,5,10,1]}\'',
                '-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_master.h5 --elem entry/data_000001 -o /scratch/detectorData/tmpOut -p \'{"sssd":[5,5,10,1]}\'',
                '-v --hdfFile /scratch/tomizaki/lysozyme1_master.h5 --elem entry/data -o /scratch/tomizaki/cbf --multiproc'
                '-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_master.h5 -o /scratch/detectorData/tmpOut',
                '-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_master.h5 -o /scratch/detectorData/tmpOut -p \'{"sssd":[2,3,4,3]}\'',
                '-v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_master.h5 -o /scratch/detectorData/tmpOut -p \'{"sssd":[2,3,4,3]}\' --multiproc',
                )

    epilog='''\n
sssd: sum skip sum divide

the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit3(0x04): cbffiles
   bit2(0x08): slicing
   bit4(0x10): ... logparse
   bit5(0x20): ... datparse
   bit6(0x40): ... mcsparse
   default is 1+2=3 (no -v)
   -v is 1+2+4=7
   -v0 is turn all off\n\n
Processing modes:
  Without -p, all images are exported to the cbf files
  -p '{"sssd":[sum0,skip,sum1,div]}'
  -p '{"osssd":[offset,sum0,skip,sum1,div]}'
  sums <sum0> images skips <skip> images and sums <sum1> of these and finally divides
  the result by <div>. e.g. -p '{"sssd":[2,3,4,1]}'
  aa...aa...aa...aa...bb...bb...bb...bb...cc...cc...cc...cc...\n\nExamples:'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    parser.add_argument('--hdfFile',       required=True, help='the hdf5 to show')
    parser.add_argument('--elem',                         help='the path to the element in the hdf5 file')
    parser.add_argument('--outDir', '-o',  required=True, help='output directory of the file(s) to generate')
    parser.add_argument('--proc', '-p',                   help='data processing mode')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=15, default=3,  help='verbose level (see below)')
    parser.add_argument('-m', '--multiproc', type=int, nargs='?', const=-1, default=0,help='use multiprocessing method topo,[numproc]')
    args = parser.parse_args()
    if args.proc==None:
      args.proc=dict()
    else:
      args.proc=json.loads(args.proc, object_hook=ut.ConvUtf8)

    LogConfig(args)
    DectrisHDF2Cbf(args).Run()
    pass

  def CheckSlicingTime(ds):
    #HDF5Slicing(ds)
    CheckSlicingTime_(ds,5,5,10)

  def HDF5Slicing(ds):
        #s=h5py.h5s.create(h5py.h5s.SEL_HYPERSLABS)
    mshape=ds.shape

    out = np.empty(mshape, dtype=ds.dtype)


    sid=ds.id.get_space()
    ds.id.read(h5py.h5s.ALL, h5py.h5s.ALL, out)
    print out[10,500:520,220]

    mshape=(10,)+mshape[1:]
    out = np.empty(mshape, dtype=ds.dtype)

    sid=ds.id.get_space()
    sid.select_hyperslab((0,0,0),mshape,op=h5py.h5s.SELECT_SET)
    print sid.get_select_bounds()

    sid_out= h5py.h5s.create_simple(mshape)
    out[:]=111
    ds.id.read(sid_out, sid, out)
    print out[0,500:520,220]
    out[:]=111
    ds.id.read(h5py.h5s.ALL, sid, out)
    print out[0,500:520,220]

    mshape=mshape[1:]
    sid.select_hyperslab((0,0,0),(5,)+mshape,op=h5py.h5s.SELECT_SET)
    sid.select_hyperslab((7,0,0),(3,)+mshape,op=h5py.h5s.SELECT_OR)
    print sid.get_select_bounds()
    out[:]=111
    ds.id.read(h5py.h5s.ALL, sid, out)
    print out[0:,500:520,220]

    # start count stride block
    #The start array specifies the offset of the starting element of the specified hyperslab.

    #The stride array chooses array locations from the dataspace with each value in the stride array
    #determining how many elements to move in each dimension. Setting a value in the stride array to 1
    #moves to each element in that dimension of the dataspace; setting a value of 2 in allocation in the
    #stride array moves to every other element in that dimension of the dataspace. In other words, the
    #stride determines the number of elements to move from the start location in each dimension. Stride
    #values of 0 are not allowed. If the stride parameter is NULL, a contiguous hyperslab is selected (as if
    #each value in the stride array were set to 1).

    #The count array determines how many blocks to select from the dataspace, in each dimension.

    #The block array determines the size of the element block selected from the dataspace. If the block
    #parameter is set to NULL, the block size defaults to a single element in each dimension (as if each value in
    #the block array were set to 1).

    #start=(0,0,0)
    #count=(5,)+mshape
    #stride=None #(1,1,1)
    #block=None  #(1,1,1)

    start=(0,0,0)
    count=(1,1,1)
    stride=(1,1,1)
    block=(2,)+mshape

    sid.select_hyperslab(start, count, stride, block)
    print sid.get_select_bounds()
    out[:]=111
    ds.id.read(h5py.h5s.ALL, sid, out)
    print out[0:,500:520,220]
    pass

  def CheckSlicingTime_(ds,nSum0,nSkip,nSum1):
    step=(nSum0+nSkip)*nSum1

    sw=ut.StopWatch()

    sw.Start()
    sl=np.arange(0,nSum0,dtype=np.uint32)
    for i in range(nSum1):
      sl1=sl+i*(nSum0+nSkip)
      v=ds[sl1,...]
      print v.shape,sl1
    sw.Log('AAA', True) #0.039613



    sw.Start()
    v=ds[0:step,...]
    print v.shape
    sw.Log('A', True) #0.129474

    sl=np.arange(0,step,dtype=np.uint32)
    sw.Start()
    v=ds[sl,...]
    print v.shape,sl
    sw.Log('B', True) #0.138759

    sl=np.arange(0,step,2,dtype=np.uint32)
    sw.Start()
    v=ds[sl,...]
    print v.shape,sl
    sw.Log('C', True) #6.505294

    sl=np.arange(0,step,dtype=np.uint32)
    sw.Start()
    v=ds[sl,...]
    w=v[0::2,...]
    print v.shape,w.shape,sl[0::2]
    sw.Log('D', True) #0.129420

    sl=(
        np.arange(0,nSum0,dtype=np.uint32).reshape(1,-1)+
        np.arange(0,nSum1,dtype=np.uint32).reshape(-1,1)*(nSum0+nSkip)
       ).ravel()
    sw.Start()
    v=ds[sl,...]
    print v.shape,sl
    sw.Log('E', True) #6.550674

    sl1=np.arange(0,step,dtype=np.uint32)
    sw.Start()
    v=ds[sl1,...]
    w=v[sl,...]
    print v.shape,w.shape,sl
    sw.Log('F', True) #0.710436

    sw.Start()
    sl=np.arange(0,step,nSum0+nSkip,dtype=np.uint32)
    for i in range(nSum0):
      sl1=sl+i
      v=ds[sl1,...]
      print v.shape,sl1
    sw.Log('G', True) #6.546652

    sw.Start()
    sl=np.arange(0,nSum0,dtype=np.uint32)
    for i in range(nSum1):
      sl1=sl+i*(nSum0+nSkip)
      v=ds[sl1,...]
      print v.shape,sl1
    sw.Log('H', True) #0.039613


    return


  def Test1():
    #scp  MAT/SUM2SKIP0div2SUM20/SUM2SKIP0div2/SUM2SKIP0div2_000*.cbf zamofing_t@pc9477:/scratch/detectorData/hdf52cbf/MAT
    #scp  PY/SUM2SKIP0div2TEST/SUM2SKIP0div2TEST/out_000*.cbf zamofing_t@pc9477:/scratch/detectorData/hdf52cbf/PY

    #BASE=/home/zamofing_t/Documents/prj/libDetXR/python/libDetXR
    #$BASE/DectrisHdf2Cbf.py -v --hdfFile /scratch/detectorData/hdf52cbf/x1_insu_0p00125d_0p00125s_0p01beam_master.h5 -o /scratch/detectorData/hdf52cbf/py.s2S0d2 -p '{"skip":0,"sum":2,"div":2}' --multiproc
    import glob
    from CbfFile import CbfParser
    cbfPy=CbfParser()
    cbfMl=CbfParser()
    #glb='/scratch/detectorData/hdf52cbf/PY/*.cbf'
    glb='/scratch/detectorData/hdf52cbf/NEW/*.cbf'
    fnPyLst=[]
    fnPyLst.extend(glob.iglob(glb))
    fnPyLst.sort()
    glb='/scratch/detectorData/hdf52cbf/OLD/*.cbf'
    fnMlLst=[]
    fnMlLst.extend(glob.iglob(glb))
    fnMlLst.sort()
    num=min(len(fnPyLst),len(fnMlLst))
    print 'compare %d images'%num
    for i in range(num):
      print fnPyLst[i],fnMlLst[i]
      cbfPy.Open(fnPyLst[i]);cbfPy.ParseHeader();(pyBuf,res)=cbfPy.GetBinary()
      cbfMl.Open(fnMlLst[i]);cbfMl.ParseHeader();(mlBuf,res)=cbfMl.GetBinary()
      noDiff =(pyBuf==mlBuf).all()
      print noDiff,
      print pyBuf[489,479],mlBuf[489,479]

      if not noDiff:
        import matplotlib.pyplot as plt
        from matplotlib.colors import LogNorm
        print pyBuf[549:570,450:470]
        print mlBuf[549:570,450:470]
        print pyBuf[291,568]
        print mlBuf[291,568]
        print pyBuf[489,479],mlBuf[489,479]
        plt.figure(0)
        imgPy=plt.imshow(pyBuf,norm=LogNorm(vmin=0.01, vmax=10),cmap = plt.cm.jet, interpolation='nearest')
        imgPy.cmap._init();bg=imgPy.cmap._lut[0].copy();bg[:-1]/=4
        imgPy.axes.set_axis_bgcolor(bg)
        plt.figure(1)
        imgMl=plt.imshow(mlBuf,norm=LogNorm(vmin=0.01, vmax=10),cmap = plt.cm.jet, interpolation='nearest')
        imgMl.cmap._init();bg=imgMl.cmap._lut[0].copy();bg[:-1]/=4
        imgMl.axes.set_axis_bgcolor(bg)
        plt.show()
        pass



#------------------ Main Code ----------------------------------
  #quick in mix arrays '{"sssd":[5,40,10,1]}'
  RunWithArgs()
  #Test1()