#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
converts scan-data on the cSAXS-Beamline to a hdf5 file.
the scan data is stored in various files/directory.
'''

import sys,os,time
import numpy as np
import traceback
import logging
import subprocess
import h5py
import utilities as ut
import multiprocessing as mproc
from libDetXR import LibDetXR

_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

from CbfFile import *
from DatParser import *
from FileNameGen import *
from csaxs_nexus import *

class SPEC2hdf5():
  _dateFmt='%Y-%m-%d_%H:%M:%S'
  _h5EntryPoint='entry/collection/data'
  def __init__(self):
    pass

  @staticmethod
  def GetScanGrp(gid,scanID):
    'gid is the /data/entry. if scanid!=None the group /data/entry/S[scanID] is build if nedded and returned'
    grp=h5py.Group(gid)
    if scanID!=None:
      name='S%.5d'%scanID
      if name in grp.keys():
        grp=grp[name]
      else:
        grp=grp.create_group(name)
    return grp

  def ParseSPEClog(self,scanID):
    '''parse *.log files'''
    fnBase=self.args.fnBase
    verbose=self.args.verbose
    idxSPEC=self.args.idxSPEC
    cmd="cd "+fnBase+" && find specES1/log-files -name '[^.]*.log' |sort"
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    retval = p.wait()
    fnLogLst=p.stdout.readlines()
    parser=SPECLogParser()
    if verbose &16:
      print 'search data for scanID %d in specES1_started_???.log-files...'%scanID
    for fnLog in fnLogLst:
      fnLog=os.path.join(fnBase,fnLog.rstrip())
      _log.debug(fnLog)
      parser.Open(fnLog)
      parser.Index(self.args.idxSPEC)
      cmdIdx=parser.GetCmdIdxOfScanIdx(scanID)
      if cmdIdx!=None:
        (cmdStr,scanRng)=parser.GetCmdInfoOfCmdIdx(cmdIdx,verbose)
        break
    parser.Close()
    if cmdIdx==None:
      _log.error('scanID %d not found in any specES1_started_???.log-files'%scanID )
      return (None,)*2
    return (cmdStr,scanRng)


  def ParseSPECdat(self,gid,scanRng):
    'parse *.dat files as .../e14428/specES1/dat-files/specES1_started_2012_03_14_2039.dat'
    fnBase=self.args.fnBase
    verbose=self.args.verbose
    idxSPEC=self.args.idxSPEC
    fnBaseDat=os.path.join(fnBase,'specES1/dat-files')
    parser=SPECDatParser()

    if verbose&32:
      print 'search scanID '+'-'.join(['%d'%k for k in scanRng])+' in specES1_started_*.dat-files...'
    if len(scanRng)==2:
      scanItr=range(scanRng[0],scanRng[1]+1)
    else:
      scanItr=scanRng
    for fnSPEC in os.listdir(fnBaseDat):
      if fnSPEC[-4:]!='.dat':
        continue
      fnSPEC=os.path.join(fnBaseDat,fnSPEC)
      parser.Open(fnSPEC)
      parser.Index(idxSPEC)
      for scanID in scanItr:
        res=parser.Parse(scanID,verbose)
        if res:
          grp=SPEC2hdf5.GetScanGrp(gid,scanID if len(scanItr)>1 else None)
          if self.args.append and grp.get('spec'):
            if verbose&32:
              print 'spec already esists in '+grp.name+'. nothing will be appened.'
            continue
          try:
            dSPEC=grp.create_dataset('spec',  data=parser.dataL)
          except AttributeError:
            _log.error('labeled table is empty')
          else:
            for k, v in zip(parser.outVar,parser.scan['pos']):
              dSPEC.attrs[k] = v
            dSPEC.attrs['scan'] = ' '.join(map(str,parser.scan['scan']))
            dSPEC.attrs['date'] = time.strftime(SPEC2hdf5._dateFmt,parser.scan['date'])
            for k,v in parser.meta.iteritems():
              dSPEC.attrs[k]=v
        else:
          break
      if res:
        break
    if not res:
      _log.error('scanID %d not found in any specES1_started_???.dat-files'%scanID )
    parser.Close()
    return res


  def ParseMCS(self,gid,scanRng):
    'parse mcs-dat files as .../e14428/mcs/S05000-05999/S05632/e14428_05632.dat'
    fnBase=self.args.fnBase
    verbose=self.args.verbose
    parser=MCSParser()
    if verbose&64:
      print 'search MCS files for scanID '+'-'.join(['%d'%k for k in scanRng])+'...'
    if len(scanRng)==2:
      scanItr=range(scanRng[0],scanRng[1]+1)
    else:
      scanItr=scanRng

    for scanID in scanItr:
      fnDirMCS=os.path.join(fnBase,'mcs','S%.5d-%.5d/S%.5d'%(scanID/1000*1000,scanID/1000*1000+999,scanID))
      ptrn='_%.5d.dat'%scanID
      fnMCS=None
      try:
        for fn in os.listdir(fnDirMCS):
          if fn.endswith(ptrn):
            fnMCS=os.path.join(fnDirMCS,fn)
            break
      except OSError: pass
      if fnMCS==None:
        if verbose&64:
          _log.warning('file '+fnDirMCS+'*'+ptrn+' not existing')
        continue
      if verbose&64:
        print 'Parse MCS file '+fnMCS

      parser.Open(fnMCS)
      parser.Parse()

      try:
        hdr=parser.header
        dataMCS=parser.dataMCS
      except AttributeError as e:
        pass
      else:
        grp=SPEC2hdf5.GetScanGrp(gid,scanID if len(scanItr)>1 else None)
        if self.args.append and grp.get('mcs'):
          if verbose&64:
            print 'mcs already esists in '+grp.name+'. nothing will be appened.'
          continue
        dMCS=grp.create_dataset('mcs',  data=dataMCS)
        for (k,v) in hdr.iteritems():
          if k=='date':
            v = time.strftime(SPEC2hdf5._dateFmt,hdr['date'])
          dMCS.attrs[k]=v

  def ParseOMNY(self,gid,scanRng):
    'parse omni/omny-dat files as ...specES1/omni/omniscan_scan_00015.dat or specES1/omny/omny_rt_scan_00016.dat'
    fnBase=self.args.fnBase
    verbose=self.args.verbose
    parser=OMNIDatParser()
    if verbose&64:
      print 'search OMNI/OMNY files for scanID '+'-'.join(['%d'%k for k in scanRng])+'...'
    if len(scanRng)==2:
      scanItr=range(scanRng[0],scanRng[1]+1)
    else:
      scanItr=scanRng

    for scanID in scanItr:
      for (key,fnPat)in(('omni','specES1/omni/omniscan_scan_%.5d.dat'),('omny','specES1/omny/omny_rt_scan_%.5d.dat')):
        fn=os.path.join(fnBase,fnPat%(scanID))
        if not os.path.exists(fn):
          if verbose&64:
            _log.warning('file '+fn+' not existing')
          continue
        if verbose&64:
          print 'Parse '+ key+' file '+fn

        parser.Open(fn)
        parser.Parse()
        try:
          hdr=parser.header
          dataOMNI=parser.dataOMNI
        except AttributeError as e:
          pass
        else:
          grp=SPEC2hdf5.GetScanGrp(gid,scanID if len(scanItr)>1 else None)
          if self.args.append and grp.get(key):
            if verbose&64:
              print key+' already esists in '+grp.name+'. nothing will be appened.'
            continue
          dMCS=grp.create_dataset(key,  data=dataOMNI)
          for (k,v) in hdr.iteritems():
            if k=='date':
              v = time.strftime(SPEC2hdf5._dateFmt,hdr['date'])
            dMCS.attrs[k]=v


  @staticmethod
  def AddCbfAttr(ds,attr):
    for k in CbfParser._keywords.keys():
      try:
        v=attr[k]
      except KeyError as e:
        continue
      if type(v)==tuple:
        if len(v)==2:
          ds.attrs[k]=v[0]
          ds.attrs['_u_'+k]=v[1]
        else:
          ds.attrs[k]=' '.join(map(str,v))
      else:
        ds.attrs[k]=v

  @staticmethod
  def AddFrameInfo(gid,instr,dateLst):
    d=dateLst[0]
    t0=(time.mktime(d[0]),d[1])
    arr=np.ndarray((len(dateLst),),np.int32)
    for idx,d in enumerate(dateLst):
      arr[idx]=(time.mktime(d[0])-t0[0])*1000+(d[1]-t0[1])
    #the timeoffset must be stored in a dataset, because the length of
    #attribute is llimited to something like 16370... and there could be more images
    grp=h5py.Group(gid)
    ds=grp.create_dataset(instr+'_info',  data=arr, compression='gzip', compression_opts=4)
    ds.attrs['startTime']=time.strftime(SPEC2hdf5._dateFmt,d[0])+'.%.1d'%d[1]
    ds.attrs['ofsTimeUnit']='ms'

  def CBF2HDF5(self,gid,dsName,fnBase,fnCBFLst):
    verbose=self.args.verbose
    cmpr=self.args.cmpr
    cbfParser=CbfParser()
    dateLst=[]
    if cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)

    for idx,fnCBF in enumerate(fnCBFLst):
      if verbose&8:
        print idx,fnCBF
      elif verbose&2:
        sys.stdout.write('.');sys.stdout.flush()
      fnCBF=os.path.join(fnBase,fnCBF)
      try:
        cbfParser.Open(fnCBF)
      except (OSError,ValueError) as e:
        traceback.print_exc(1)
        continue
      try:
        cbfParser.ParseHeader(7)
      except BaseException as e:
        traceback.print_exc(1)
        continue
      (outBuf,res)=cbfParser.GetBinary()
      attr=cbfParser.header['header_contents']
      if idx==0:
        dtype=outBuf.dtype
        shape=(len(fnCBFLst),)+outBuf.shape
        maxshape=(h5py.h5s.UNLIMITED,)+outBuf.shape
        chunk=(1,)+outBuf.shape
        dsid=ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,cmpr)
        ds=h5py.Dataset(dsid)
        SPEC2hdf5.AddCbfAttr(ds,attr)
        _log.debug(' '.join(map(str,(outBuf.shape,ds.shape,cbfParser.header))))
      try:
        d=attr['date']
      except KeyError as e:
        pass
      else:
        dateLst.append(d)
      assert(outBuf.flags.c_contiguous)
      if cmpr:
        (cmprBuf,filter)=CmprChunkBuild(outBuf)
        dsid.write_chunk((idx,0,0),cmprBuf,filters=filter)
      else:
        dsid.write_chunk((idx,0,0),outBuf)
      cbfParser.Close()
    if verbose&2 and not verbose&8:
      sys.stdout.write('\n') #newline after all the .
    SPEC2hdf5.AddFrameInfo(gid,dsName,dateLst)
    return ds

  @staticmethod
  def MultiProcChunkBuilder(ChunkBuildFunc,pipe):
    cbfParser=CbfParser()
    #print 'running MultiProcChunkBuilder PID %d'%os.getpid()
    while True:
      fnCBF=pipe.recv()
      if fnCBF==None:
        break
      try:
        cbfParser.Open(fnCBF)
      except (OSError,ValueError) as e:
        traceback.print_exc(1)
        pipe.send(None)
        continue
      try:
        cbfParser.ParseHeader(7)
      except BaseException as e:
        traceback.print_exc(1)
        pipe.send(None)
        continue
      (outBuf,res)=cbfParser.GetBinary()
      attr=cbfParser.header['header_contents']
      (cmprBuf,filter)=ChunkBuildFunc(outBuf)
      cbfParser.Close()
      pipe.send((cmprBuf,filter,attr,outBuf.dtype,outBuf.shape))
    #print 'end MultiProcChunkBuilder PID %d'%os.getpid()
    return 0

  def MultiProcCBF2HDF5(self,gid,dsName,fnBase,fnCBFLst):
    verbose=self.args.verbose
    cmpr=self.args.cmpr
    ChunkBuildFunc=ut.GetChunkBuildFunc(cmpr)
    dateLst=[]

    jobs = []
    maxJobs=mproc.cpu_count()
    for idx in range(maxJobs):
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=SPEC2hdf5.MultiProcChunkBuilder, args=(ChunkBuildFunc,pipeChild))
      jobs.append((p,pipeParent))
      p.start()
    num=len(fnCBFLst)
    for idx in range(num+maxJobs):
      (p,pipeParent)=jobs[idx%maxJobs]
      if idx>=maxJobs: #receive compressed data
        idxRes=idx-maxJobs
        (cmprBuf,filter,attr,dtype,shape)=pipeParent.recv()
        if idxRes==0:
          #dtype=outBuf.dtype
          #shape=(num,)+outBuf.shape
          #maxshape=(h5py.h5s.UNLIMITED,)+outBuf.shape
          chunk=(1,)+shape
          dsshape=(num,)+shape
          maxshape=(h5py.h5s.UNLIMITED,)+shape
          chunk=(1,)+shape
          dsid=ut.CreateDatasetChunked(gid,dsName,dtype,dsshape,maxshape,chunk,cmpr)
          ds=h5py.Dataset(dsid)
          SPEC2hdf5.AddCbfAttr(ds,attr)
        try:
          d=attr['date']
        except KeyError as e:
          pass
        else:
          dateLst.append(d)
        if verbose&4 and not verbose&8:
          sys.stdout.write('w');sys.stdout.flush()
        dsid.write_chunk((idxRes,0,0),cmprBuf,filters=filter)
      if idx<num:
        fnCBF=fnCBFLst[idx]
        if verbose&8:
          print idx,fnCBF
        elif verbose &(2+4):
          sys.stdout.write('.');sys.stdout.flush()
        fnCBF=os.path.join(fnBase,fnCBF)
      else:
        fnCBF=None
      pipeParent.send(fnCBF)

    if verbose&(2+4) and not verbose&8:
      sys.stdout.write('\n') #newline after all the .
    SPEC2hdf5.AddFrameInfo(gid,dsName,dateLst)
    return ds

  @staticmethod
  def GetCbfFileList(scanItr,fnBaseDet,verbose):
    fnCBFLst=[]
    for scanID in scanItr:
      d='S%.5d-%.5d/S%.5d'%(scanID/1000*1000,scanID/1000*1000+999,scanID)
      fnBaseCBF=os.path.join(fnBaseDet,d)
      try:
        for cbfFile in os.listdir(fnBaseCBF):
          #print cbfFile
          if cbfFile[-4:]=='.cbf':
            fnCBFLst.append(os.path.join(d,cbfFile))
      except OSError:
        break
    if not fnCBFLst:
      return None
    fnCBFLst.sort()
    if verbose&1:
      print 'append %d CBF-files in path %s'%(len(fnCBFLst),fnBaseDet)
    return fnCBFLst

  def Run(self,args):
    self.args=args
    fnBase=args.fnBase
    verbose=args.verbose
    scanRng=args.scanID
    grpScan=args.grpScan
    mode='a' if args.append else 'w'

    ut.StopWatch.Start()
    if grpScan:
      if type(scanRng)!=int:
        _log.error('with option grpScan scanID-range is not allowed. grpScan ignored.')
        grpScan=args.grpScan=False
      else:
        (cmdStr,scanRng)=self.ParseSPEClog(scanRng)
        if scanRng==None: return
    if type(scanRng)!=tuple:
      scanRng=(scanRng,)
    if verbose&2:
      ut.StopWatch.Log('ParseLog', True)
    outDir=args.outDir
    if outDir==None:
      outDir=fnBase

    fnHdf5=os.path.join(outDir,'scan_'+'-'.join(['%.5d'%k for k in scanRng])+'.hdf5')

    try:
      fid=ut.OpenFile(fnHdf5,mode)
    except IOError as e:
      _log.error(fnHdf5+'\n'+str(e))
      return
    if grpScan:
      obj=h5py.File(fid)
      obj.attrs['SPEC command'] = cmdStr
    res,v=LibDetXR.GetVersion()
    attr=(('creator','Converted data with '+os.path.basename(__file__)+' ver. %d.%d.%d.%d'%tuple(v[::-1])+' from data scan '+'-'.join(['%.d'%k for k in scanRng])),
          ('instrument',"My Used Instruments..."),
          ('NeXus_version',"?.?.?"))
    ut.AddAttributes(fid,attr)
    gid=ut.CreateGrpPath(fid,SPEC2hdf5._h5EntryPoint)
    self.ParseSPECdat(gid,scanRng)
    if verbose&2:
      ut.StopWatch.Log('ParseSPEC', True)
    self.ParseMCS(gid,scanRng)
    if verbose&2:
      ut.StopWatch.Log('ParseMCS', True)
    self.ParseOMNY(gid,scanRng)
    if verbose&2:
      ut.StopWatch.Log('ParseOMNY', True)

    if len(scanRng)==2:
      scanItr=range(scanRng[0],scanRng[1]+1)
    else:
      scanItr=scanRng
    for detStr in os.listdir(fnBase):
      fnBaseDet=os.path.join(fnBase,detStr)
      if not os.path.isdir(fnBaseDet):
        continue
      fnCBFLst=SPEC2hdf5.GetCbfFileList(scanItr,fnBaseDet,verbose)
      if not fnCBFLst:
        continue
        #OR:
        #fng=FileNameGen()
        #fng.SetPatternVarRange(fnBaseCBF+'/%s_%d_%.5d_%.5d_%.5d.cbf',
        #'account,detID,scanID,X,Y',(('account',(account,)),('detID',(detID,)),('scanID',(scanID,)),('X',10),('Y',50)))
        #
        #for p in range(10*50):
        #  print fng.GetFileName(0, p)
        #pass
      if args.append and gid.links.exists(detStr):
        if verbose&32:
          print detStr+' already esists in '+h5py.Group(gid).name+'. nothing will be appened.'
        continue #ignore existing acquisition data
      if args.cmpr and args.multiproc:
        ds=self.MultiProcCBF2HDF5(gid,detStr,fnBaseDet,fnCBFLst)
      else:
        ds=self.CBF2HDF5(gid,detStr,fnBaseDet,fnCBFLst)

    #adding external links to other hdf5 files
    for detStr in ('pilatus_1','pilatus_2','eiger_1'):
      fnDet=detStr+'_'+'-'.join(['%.5d'%k for k in scanRng])+'.hdf5'
      if os.path.isfile(os.path.join(fnBase,fnDet)):
        try:
          gid.links.create_external(detStr,fnDet,SPEC2hdf5._h5EntryPoint+'/'+detStr)
          if verbose&1:
            print 'append external link from: '+fnDet+SPEC2hdf5._h5EntryPoint+'/'+detStr
        except RuntimeError as e:
          _log.error(str(e)+'\nExternal Link: '+' '.join((detStr,fnDet,SPEC2hdf5._h5EntryPoint+'/'+detStr)))

    if args.nexus:
      CSAXS_Nexus(fid).AddStructure()
    ut.CloseFile(fid)
    if verbose&2:
      ut.StopWatch.Log('CBF2HDF5', True)

def RunWithArgs(args):
  hdlr=logging.StreamHandler()
  fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
  hdlr.setFormatter(fmt)
  del logging.root.handlers[:]
  logging.root.addHandler(hdlr)
  logging.root.setLevel(logging.DEBUG)

  logLvls={'CbfFile'    : logging.WARN,
           'DatParser'  : logging.WARN,
           'FileNameGen': logging.WARN,
           'libDetXR'   : logging.WARN,
           __name__     : logging.WARN,
          }
  for k,v in logLvls.iteritems():
    logging.getLogger(k).setLevel(v)
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
             'fnBase':'~/Data10',
             'outDir':None,
             'scanID':None,
             'grpScan':False,
             'verbose':3,
             'idxSPEC':7,
             'multiproc':False,
             'cmpr':None,
             'append':False,
             'nexus':False
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
  SPEC2hdf5().Run(args)

if __name__ == '__main__':
  import argparse #since python 2.7
  def ParseArgs():

    if len(sys.argv)>1 and sys.argv[1][:4]=='Test':
      print eval(sys.argv[1]+'()')
      return

    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--baseDir=/scratch/detectorData/e14472 33',
                '-v115 --baseDir=/scratch/detectorData/e14472 33 --grpScan --cmpr s-zlib4 --multiproc',
                '-v --baseDir=/scratch/detectorData/stxm_pinhole --outDir=/tmp 158-198 --cmpr s-zlib4 --multiproc',
                )
    epilog='''\n
the verbose bits are:
   bit0( 1): basic
   bit1( 2): progress
   bit2( 4): multiproc
   bit3( 8): cbffiles
   bit4(16): logparse
   bit5(32): datparse
   bit6(64): mcsparse
   default is 1+2=3 (no -v)
   -v is 1+2+4+16+32+64=119
   -v0 is turn all off

 indexing:
   The .dat and .log files are indexed to locate the scan data in these huge text files
   - ignore: ignore existing idx file, do indexing, dont write idx-file
   - force:  ignore existing idx file, do indexing, overwrite idx file
   - update: use existing idx file, do indexing, update existing idx file

Some of these files look like (example for scan 33 in dir /sls/X12SA/data/e14472/Data10):
 - specES1/dat-files/specES1_started_2013_04_26_1604.dat
 - mcs/S00000-00999/S00033/e14472_00033.dat
 - eiger/S00000-00999/S00033/e14472_1_00033_00000_00001.cbf and many more
 - pilatus_1/S00000-00999/S00033/e14472_1_00033_00000_00001.cbf and many more
 - pilatus_2/S00000-00999/S00033/e14472_1_00033_00000_00001.cbf and many more

All these files (can be thousands for one scan) are stores in a single scan file:
 - e14472_00033.hdf5

Example:'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    idxSPEC={'ignore':2,'update':7,'force':6}

    parser.add_argument('--baseDir',  dest='fnBase',   default='~/Data10',  help='base directory where specES1, mcs, pilatus_1 etc. is located')
    parser.add_argument('--outDir' ,                                        help='output directory of the file(s) to generate (default:same as BaseDir')
    parser.add_argument('scanID',                                           help='the scan number to store in the hdf5 file')
    parser.add_argument('--grpScan',                   action='store_true', help='groups multiple scans of one command in one hdf5 file')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=119,default=3,  help='verbose level (see below)')

    parser.add_argument('--idx',      dest='idxSPEC',  choices=idxSPEC, default='update', help='indexing SPEC.dat and SPEC.log files (default:update)')

    parser.add_argument('--multiproc',                 action='store_true', help='use multiprocessing method')
    parser.add_argument('--cmpr',                                           help='compression method (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('--append', '-a',              action='store_true', help='appends SPEC to existing file. Existing objects are not affected. ')
    parser.add_argument('--nexus', '-n',               action='store_true', help='try to add nexus complient structure')
    args = parser.parse_args()

    args.shuffle=True if args.cmpr else False
    args.idxSPEC=idxSPEC[args.idxSPEC]

    try:
      args.scanID=int(args.scanID)
    except ValueError:
      try:
        args.scanID=tuple(map(int,args.scanID.split('-')))
      except ValueError:
        parser.error('scanID: cant parse '+args.scanID)

    if not os.path.exists(args.fnBase):
      args.fnBase=os.path.expanduser(args.fnBase)
    if args.outDir and not os.path.exists(args.outDir):
      args.outDir=os.path.expanduser(args.outDir)

    RunWithArgs(args)
    pass

#------------------ Main Code ----------------------------------
  ParseArgs()
